# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=too-many-public-methods
# pylint: disable-msg=broad-except
# pylint: disable-msg=import-error
"""Qiskit Metal unit tests components functionality."""

import unittest
import numpy as np

from qiskit_metal.qlibrary.core import QComponent
from qiskit_metal.qlibrary.core import QRoute
from qiskit_metal.qlibrary.core import QRouteLead
from qiskit_metal.qlibrary.core import QRoutePoint
from qiskit_metal.qlibrary.core import BaseQubit
from qiskit_metal.qlibrary.sample_shapes.circle_caterpillar import CircleCaterpillar
from qiskit_metal.qlibrary.sample_shapes.circle_raster import CircleRaster
from qiskit_metal.qlibrary.sample_shapes.rectangle import Rectangle
from qiskit_metal.qlibrary.sample_shapes.rectangle_hollow import RectangleHollow
from qiskit_metal.qlibrary.sample_shapes.n_gon import NGon
from qiskit_metal.qlibrary.sample_shapes.n_square_spiral import NSquareSpiral
from qiskit_metal.qlibrary.lumped.cap_n_interdigital import CapNInterdigital
from qiskit_metal.qlibrary.couplers.coupled_line_tee import CoupledLineTee
from qiskit_metal.qlibrary.couplers.cap_n_interdigital_tee import CapNInterdigitalTee
from qiskit_metal.qlibrary.couplers.line_tee import LineTee
from qiskit_metal.qlibrary.terminations.open_to_ground import OpenToGround
from qiskit_metal.qlibrary.terminations.short_to_ground import ShortToGround
from qiskit_metal.qlibrary.tlines.anchored_path import RouteAnchors
from qiskit_metal.qlibrary.tlines.framed_path import RouteFramed
from qiskit_metal.qlibrary.tlines.straight_path import RouteStraight
from qiskit_metal.qlibrary.tlines.pathfinder import RoutePathfinder
from qiskit_metal.qlibrary.tlines.meandered import RouteMeander
from qiskit_metal.qlibrary.tlines.mixed_path import RouteMixed
from qiskit_metal.qlibrary.terminations.launchpad_wb import LaunchpadWirebond
from qiskit_metal.qlibrary.terminations.launchpad_wb_coupled import LaunchpadWirebondCoupled
from qiskit_metal.qlibrary.lumped.cap_3_interdigital import Cap3Interdigital
from qiskit_metal.qlibrary.qubits.star_qubit import StarQubit
from qiskit_metal.qlibrary.qubits.JJ_Dolan import jj_dolan
from qiskit_metal.qlibrary.qubits.JJ_Manhattan import jj_manhattan
from qiskit_metal.qlibrary.qubits.transmon_concentric import TransmonConcentric
from qiskit_metal.qlibrary.qubits.transmon_cross import TransmonCross
from qiskit_metal.qlibrary.qubits.transmon_cross_fl import TransmonCrossFL
from qiskit_metal.qlibrary.qubits.transmon_pocket import TransmonPocket
from qiskit_metal.qlibrary.qubits.transmon_pocket_cl import TransmonPocketCL
from qiskit_metal.qlibrary.qubits.transmon_pocket_6 import TransmonPocket6
from qiskit_metal.qlibrary.qubits.transmon_pocket_teeth import TransmonPocketTeeth
from qiskit_metal.qlibrary.qubits.SQUID_loop import SQUID_LOOP
from qiskit_metal.qlibrary.couplers.tunable_coupler_01 import TunableCoupler01
from qiskit_metal import designs
from qiskit_metal.qlibrary._template import MyQComponent
from qiskit_metal.tests.assertions import AssertionsMixin

#pylint: disable-msg=line-too-long
from qiskit_metal.qlibrary.lumped.resonator_coil_rect import ResonatorCoilRect


class TestComponentInstantiation(unittest.TestCase, AssertionsMixin):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_qlibrary_instantiate_qcomponent(self):
        """Test the instantiaion of QComponent."""
        design = designs.DesignPlanar()
        try:
            QComponent
        except Exception:
            self.fail("QComponent failed")

        with self.assertRaises(NotImplementedError):
            QComponent(design, "my_name")

        with self.assertRaises(NotImplementedError):
            QComponent(design, "my_name2", options={})

        try:
            QComponent(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "QComponent(design, \"my_name3\", options={}, make=False)")

        try:
            QComponent(design,
                       "my_name4",
                       options={},
                       make=False,
                       component_template={})
        except Exception:
            msg = "QComponent(design, \"my_name4\", options={}, make=False, component_template={})"
            self.fail(msg)

    def test_qlibrary_instantiate_basequbit(self):
        """Test the instantiation of basequbit."""
        design = designs.DesignPlanar()
        try:
            BaseQubit
        except Exception:
            self.fail("BaseQubit failed")

        with self.assertRaises(NotImplementedError):
            BaseQubit(design, "my_name")

        with self.assertRaises(NotImplementedError):
            BaseQubit(design, "my_name2", options={})

        try:
            BaseQubit(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail("BaseQubit(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_open_to_ground(self):
        """Test the instantiation of openToGround."""
        design = designs.DesignPlanar()
        try:
            OpenToGround
        except Exception:
            self.fail("OpenToGround failed")

        try:
            OpenToGround(design, "my_name")
        except Exception:
            self.fail("OpenToGround(design, \"my_name\")")

        try:
            OpenToGround(design, "my_name2", options={})
        except Exception:
            self.fail("OpenToGround(design, \"my_name2\", options={})")

        try:
            OpenToGround(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "OpenToGround(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_short_to_ground(self):
        """Test the instantiation of shortToGround."""
        design = designs.DesignPlanar()
        try:
            ShortToGround
        except Exception:
            self.fail("ShortToGround failed")

        try:
            ShortToGround(design, "my_name")
        except Exception:
            self.fail("ShortToGround(design, \"my_name\")")

        try:
            ShortToGround(design, "my_name2", options={})
        except Exception:
            self.fail("ShortToGround(design, \"my_name2\", options={})")

        try:
            ShortToGround(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "ShortToGround(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_route_frame_path(self):
        """Test the instantiation of RouteFramed."""
        try:
            RouteFramed
        except Exception:
            self.fail("RouteFramed failed")

    def test_qlibrary_instantiate_route_straight(self):
        """Test the instantiation of RouteStraight."""
        design = designs.DesignPlanar()
        try:
            RouteStraight
        except Exception:
            self.fail("RouteStraight failed")

        try:
            RouteStraight(design, "my_name2", options={})
        except Exception:
            self.fail("RouteStraight(design, \"my_name2\", options={})")

        try:
            RouteStraight(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "RouteStraight(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_q_route_lead(self):
        """Test the instantiation of QRouteLead."""
        try:
            QRouteLead
        except Exception:
            self.fail("QRouteLead failed")

    def test_qlibrary_instantiate_q_route_point(self):
        """Test the instantiation of QRoutePoint."""
        try:
            QRoutePoint(np.array([1, 2]))
        except Exception:
            self.fail("QRoutePoint failed")

    def test_qlibrary_instantiate_route_meander(self):
        """Test the instantiation of RouteMeander."""
        try:
            RouteMeander
        except Exception:
            self.fail("RouteMeander failed")

    def test_qlibrary_instantiate_route_mixed(self):
        """Test the instantiation of RouteMixed."""
        try:
            RouteMixed
        except Exception:
            self.fail("RouteMixed failed")

    def test_qlibrary_instantiate_q_route(self):
        """Test the instantiation of QRoute."""
        design = designs.DesignPlanar()
        try:
            QRoute(design, name='test_qroute', options={})
        except Exception:
            self.fail("QRoute failed")

    def test_qlibrary_instantiate_my_q_component(self):
        """Test the instantiation of MyQComponent."""
        design = designs.DesignPlanar()
        try:
            MyQComponent
        except Exception:
            self.fail("MyQComponent failed")

        try:
            MyQComponent(design, "my_name")
        except Exception:
            self.fail("MyQComponent(design, \"my_name\")")

        try:
            MyQComponent(design, "my_name2", options={})
        except Exception:
            self.fail("MyQComponent(design, \"my_name2\", options={})")

        try:
            MyQComponent(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "MyQComponent(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_cpw_finger_cap(self):
        """Test the instantiation of CapNInterdigital."""
        design = designs.DesignPlanar()
        try:
            CapNInterdigital
        except Exception:
            self.fail("CapNInterdigital failed")

        try:
            CapNInterdigital(design, "my_name")
        except Exception:
            self.fail("CapNInterdigital(design, \"my_name\")")

        try:
            CapNInterdigital(design, "my_name2", options={})
        except Exception:
            self.fail("CapNInterdigital(design, \"my_name2\", options={})")

        try:
            CapNInterdigital(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "CapNInterdigital(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_instantiate_cpw_t_finger_cap(self):
        """Test the instantiation of CapNInterdigitalTee."""
        design = designs.DesignPlanar()
        try:
            CapNInterdigitalTee
        except Exception:
            self.fail("CapNInterdigitalTee failed")

        try:
            CapNInterdigitalTee(design, "my_name")
        except Exception:
            self.fail("CapNInterdigitalTee(design, \"my_name\")")

        try:
            CapNInterdigitalTee(design, "my_name2", options={})
        except Exception:
            self.fail("CapNInterdigitalTee(design, \"my_name2\", options={})")

        try:
            CapNInterdigitalTee(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "CapNInterdigitalTee(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_instantiate_cpw_t(self):
        """Test the instantiation of LineTee."""
        design = designs.DesignPlanar()
        try:
            LineTee
        except Exception:
            self.fail("LineTee failed")

        try:
            LineTee(design, "my_name")
        except Exception:
            self.fail("LineTee(design, \"my_name\")")

        try:
            LineTee(design, "my_name2", options={})
        except Exception:
            self.fail("LineTee(design, \"my_name2\", options={})")

        try:
            LineTee(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail("LineTee(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_cpw_hanger_t(self):
        """Test the instantiation of CoupledLineTee."""
        design = designs.DesignPlanar()
        try:
            CoupledLineTee
        except Exception:
            self.fail("CoupledLineTee failed")

        try:
            CoupledLineTee(design, "my_name")
        except Exception:
            self.fail("CoupledLineTee(design, \"my_name\")")

        try:
            CoupledLineTee(design, "my_name2", options={})
        except Exception:
            self.fail("CoupledLineTee(design, \"my_name2\", options={})")

        try:
            CoupledLineTee(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "CoupledLineTee(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_resonator_rectangle_spiral(self):
        """Test the instantiation of ResonatorCoilRect."""
        design = designs.DesignPlanar()
        try:
            ResonatorCoilRect
        except Exception:
            self.fail("ResonatorCoilRect failed")

        try:
            ResonatorCoilRect(design, "my_name")
        except Exception:
            self.fail("ResonatorCoilRect(design, \"my_name\")")

        try:
            ResonatorCoilRect(design, "my_name2", options={})
        except Exception:
            self.fail("ResonatorCoilRect(design, \"my_name2\", options={})")

        try:
            ResonatorCoilRect(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "ResonatorCoilRect(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_instantiate_circle_raster(self):
        """Test the instantiation of CircleRaster."""
        design = designs.DesignPlanar()
        try:
            CircleRaster
        except Exception:
            self.fail("CircleRaster failed")

        try:
            CircleRaster(design, "my_name")
        except Exception:
            self.fail("CircleRaster(design, \"my_name\")")

        try:
            CircleRaster(design, "my_name2", options={})
        except Exception:
            self.fail("CircleRaster(design, \"my_name2\", options={})")

        try:
            CircleRaster(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "CircleRaster(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_circle_caterpillar(self):
        """Test the instantiation of CircleCaterpillar."""
        design = designs.DesignPlanar()
        try:
            CircleCaterpillar
        except Exception:
            self.fail("CircleCaterpillar failed")

        try:
            CircleCaterpillar(design, "my_name")
        except Exception:
            self.fail("CircleCaterpillar(design, \"my_name\")")

        try:
            CircleCaterpillar(design, "my_name2", options={})
        except Exception:
            self.fail("CircleCaterpillar(design, \"my_name2\", options={})")

        try:
            CircleCaterpillar(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "CircleCaterpillar(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_instantiate_n_gon(self):
        """Test the instantiation of NGon."""
        design = designs.DesignPlanar()
        try:
            NGon
        except Exception:
            self.fail("NGon failed")

        try:
            NGon(design, "my_name")
        except Exception:
            self.fail("NGon(design, \"my_name\")")

        try:
            NGon(design, "my_name2", options={})
        except Exception:
            self.fail("NGon(design, \"my_name2\", options={})")

        try:
            NGon(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail("NGon(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_n_square_spiral(self):
        """Test the instantiation of NSquareSpiral."""
        design = designs.DesignPlanar()
        try:
            NSquareSpiral
        except Exception:
            self.fail("NSquareSpiral failed")

        try:
            NSquareSpiral(design, "my_name")
        except Exception:
            self.fail("NSquareSpiral(design, \"my_name\")")

        try:
            NSquareSpiral(design, "my_name2", options={})
        except Exception:
            self.fail("NSquareSpiral(design, \"my_name2\", options={})")

        try:
            NSquareSpiral(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "NSquareSpiral(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_rectangle(self):
        """Test the instantiation of Rectangle."""
        design = designs.DesignPlanar()
        try:
            Rectangle
        except Exception:
            self.fail("Rectangle failed")

        try:
            Rectangle(design, "my_name")
        except Exception:
            self.fail("Rectangle(design, \"my_name\")")

        try:
            Rectangle(design, "my_name2", options={})
        except Exception:
            self.fail("Rectangle(design, \"my_name2\", options={})")

        try:
            Rectangle(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail("Rectangle(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_rectangle_hollow(self):
        """Test the instantiation of RectangleHollow."""
        design = designs.DesignPlanar()
        try:
            RectangleHollow
        except Exception:
            self.fail("RectangleHollow failed")

        try:
            RectangleHollow(design, "my_name")
        except Exception:
            self.fail("RectangleHollow(design, \"my_name\")")

        try:
            RectangleHollow(design, "my_name2", options={})
        except Exception:
            self.fail("RectangleHollow(design, \"my_name2\", options={})")

        try:
            RectangleHollow(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "RectangleHollow(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_route_anchors(self):
        """Test the instantiation of RouteAnchors."""
        design = designs.DesignPlanar()
        try:
            RouteAnchors
        except Exception:
            self.fail("RouteAnchors failed")

        try:
            RouteAnchors(design, "my_name2", options={})
        except Exception:
            self.fail("RouteAnchors(design, \"my_name2\", options={})")

        try:
            RouteAnchors(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "RouteAnchors(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_route_pathfinder(self):
        """Test the instantiation of RoutePathfinder."""
        design = designs.DesignPlanar()
        try:
            RoutePathfinder
        except Exception:
            self.fail("RoutePathfinder failed")

        try:
            RoutePathfinder(design, "my_name2", options={})
        except Exception:
            self.fail("RoutePathfinder(design, \"my_name2\", options={})")

        try:
            RoutePathfinder(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "RoutePathfinder(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_instantiate_launch_v1(self):
        """Test the instantiation of LaunchpadWirebond."""
        design = designs.DesignPlanar()
        try:
            LaunchpadWirebond
        except Exception:
            self.fail("LaunchpadWirebond failed")

        try:
            LaunchpadWirebond(design, "my_name")
        except Exception:
            self.fail("LaunchpadWirebond(design, \"my_name\")")

        try:
            LaunchpadWirebond(design, "my_name2", options={})
        except Exception:
            self.fail("LaunchpadWirebond(design, \"my_name2\", options={})")

        try:
            LaunchpadWirebond(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "LaunchpadWirebond(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_instantiate_launch_v2(self):
        """Test the instantiation of LaunchpadWirebondCoupled."""
        design = designs.DesignPlanar()
        try:
            LaunchpadWirebondCoupled
        except Exception:
            self.fail("LaunchpadWirebondCoupled failed")

        try:
            LaunchpadWirebondCoupled(design, "my_name")
        except Exception:
            self.fail("LaunchpadWirebondCoupled(design, \"my_name\")")

        try:
            LaunchpadWirebondCoupled(design, "my_name2", options={})
        except Exception:
            self.fail(
                "LaunchpadWirebondCoupled(design, \"my_name2\", options={})")

        try:
            LaunchpadWirebondCoupled(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "LaunchpadWirebondCoupled(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_instantiate_three_finger_cap_v1(self):
        """Test the instantiation of Cap3Interdigital."""
        design = designs.DesignPlanar()
        try:
            Cap3Interdigital
        except Exception:
            self.fail("Cap3Interdigital failed")

        try:
            Cap3Interdigital(design, "my_name")
        except Exception:
            self.fail("Cap3Interdigital(design, \"my_name\")")

        try:
            Cap3Interdigital(design, "my_name2", options={})
        except Exception:
            self.fail("Cap3Interdigital(design, \"my_name2\", options={})")

        try:
            Cap3Interdigital(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "Cap3Interdigital(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_qubits_transmon_concentric(self):
        """Test the instantiation of TransmonConcentric."""
        design = designs.DesignPlanar()
        try:
            TransmonConcentric
        except Exception:
            self.fail("TransmonConcentric failed")

        try:
            TransmonConcentric(design, "my_name")
        except Exception:
            self.fail("TransmonConcentric(design, \"my_name\") failed")

        try:
            TransmonConcentric(design, "my_name2", options={})
        except Exception:
            self.fail(
                "TransmonConcentric(design, \"my_name2\", options={}) failed")

        try:
            TransmonConcentric(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonConcentric(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_qubits_star_qubit(self):
        """Test the instantiation of StarQubit."""
        design = designs.DesignPlanar()
        try:
            StarQubit
        except Exception:
            self.fail("StarQubit failed")

        try:
            StarQubit(design, "my_name")
        except Exception:
            self.fail("StarQubit(design, \"my_name\") failed")

        try:
            StarQubit(design, "my_name2", options={})
        except Exception:
            self.fail("StarQubit(design, \"my_name2\", options={}) failed")

        try:
            StarQubit(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail("StarQubit(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_transmon_cross(self):
        """Test the instantiation of TransmonCross."""
        design = designs.DesignPlanar()
        try:
            TransmonCross
        except Exception:
            self.fail("TransmonCross failed")

        try:
            TransmonCross(design, "my_name")
        except Exception:
            self.fail("TransmonCross(design, \"my_name\") failed")

        try:
            TransmonCross(design, "my_name2", options={})
        except Exception:
            self.fail("TransmonCross(design, \"my_name2\", options={}) failed")

        try:
            TransmonCross(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonCross(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_transmon_cross_fl(self):
        """Test the instantiation of TransmonCrossFL."""
        design = designs.DesignPlanar()
        try:
            TransmonCrossFL
        except Exception:
            self.fail("TransmonCrossFL failed")

        try:
            TransmonCrossFL(design, "my_name")
        except Exception:
            self.fail("TransmonCrossFL(design, \"my_name\") failed")

        try:
            TransmonCrossFL(design, "my_name2", options={})
        except Exception:
            self.fail(
                "TransmonCrossFL(design, \"my_name2\", options={}) failed")

        try:
            TransmonCrossFL(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonCrossFL(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_transmon_pocket(self):
        """Test the instantiation of TransmonPocket."""
        design = designs.DesignPlanar()
        try:
            TransmonPocket
        except Exception:
            self.fail("TransmonPocket failed")

        try:
            TransmonPocket(design, "my_name")
        except Exception:
            self.fail("TransmonPocket(design, \"my_name\") failed")

        try:
            TransmonPocket(design, "my_name2", options={})
        except Exception:
            self.fail("TransmonPocket(design, \"my_name2\", options={}) failed")

        try:
            TransmonPocket(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonPocket(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_transmon_pocket_cl(self):
        """Test the instantiation of TransmonPocketCL."""
        design = designs.DesignPlanar()
        try:
            TransmonPocketCL
        except Exception:
            self.fail("TransmonPocketCL failed")

        try:
            TransmonPocketCL(design, "my_name")
        except Exception:
            self.fail("TransmonPocketCL(design, \"my_name\") failed")

        try:
            TransmonPocketCL(design, "my_name2", options={})
        except Exception:
            self.fail(
                "TransmonPocketCL(design, \"my_name2\", options={}) failed")

        try:
            TransmonPocketCL(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonPocketCL(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_qubits_transmon_pocket_6(self):
        """Test the instantiation of TransmonPocket6."""
        design = designs.DesignPlanar()
        try:
            TransmonPocket6
        except Exception:
            self.fail("TransmonPocket6 failed")

        try:
            TransmonPocket6(design, "my_name")
        except Exception:
            self.fail("TransmonPocket6(design, \"my_name\") failed")

        try:
            TransmonPocket6(design, "my_name2", options={})
        except Exception:
            self.fail(
                "TransmonPocket6(design, \"my_name2\", options={}) failed")

        try:
            TransmonPocket6(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonPocket6(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_transmon_pocket_teeth(self):
        """Test the instantiation of TransmonPocketTeeth."""
        design = designs.DesignPlanar()
        try:
            TransmonPocketTeeth
        except Exception:
            self.fail("TransmonPocketTeeth failed")

        try:
            TransmonPocketTeeth(design, "my_name")
        except Exception:
            self.fail("TransmonPocketTeeth(design, \"my_name\") failed")

        try:
            TransmonPocketTeeth(design, "my_name2", options={})
        except Exception:
            self.fail(
                "TransmonPocketTeeth(design, \"my_name2\", options={}) failed")

        try:
            TransmonPocketTeeth(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TransmonPocketTeeth(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_qubits_squid_loop(self):
        """Test the instantiation of SQUID_LOOP."""
        design = designs.DesignPlanar()
        try:
            SQUID_LOOP
        except Exception:
            self.fail("SQUID_LOOP failed")

        try:
            SQUID_LOOP(design, "my_name")
        except Exception:
            self.fail("SQUID_LOOP(design, \"my_name\") failed")

        try:
            SQUID_LOOP(design, "my_name2", options={})
        except Exception:
            self.fail("SQUID_LOOP(design, \"my_name2\", options={}) failed")

        try:
            SQUID_LOOP(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "SQUID_LOOP(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_tunable_coupler_01(self):
        """Test the instantiation of TunableCoupler01."""
        design = designs.DesignPlanar()
        try:
            TunableCoupler01
        except Exception:
            self.fail("TunableCoupler01 failed")

        try:
            TunableCoupler01(design, "my_name")
        except Exception:
            self.fail("TunableCoupler01(design, \"my_name\") failed")

        try:
            TunableCoupler01(design, "my_name2", options={})
        except Exception:
            self.fail(
                "TunableCoupler01(design, \"my_name2\", options={}) failed")

        try:
            TunableCoupler01(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "TunableCoupler01(design, \"my_name3\", options={}, make=False)"
            )

    def test_qlibrary_qubits_jj_dolan_instantiate(self):
        """Test the instantiaion of jj_dolan."""
        design = designs.DesignPlanar()
        try:
            jj_dolan
        except Exception:
            self.fail("jj_dolan failed")

        try:
            jj_dolan(design, "my_name2", options={})
        except Exception:
            self.fail("jj_dolan(design, \"my_name2\", options={})")

        try:
            jj_dolan(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail("jj_dolan(design, \"my_name3\", options={}, make=False)")

    def test_qlibrary_qubits_jj_manhattan_instantiate(self):
        """Test the instantiaion of jj_dolan."""
        design = designs.DesignPlanar()
        try:
            jj_manhattan
        except Exception:
            self.fail("jj_manhattan failed")

        try:
            jj_manhattan(design, "my_name2", options={})
        except Exception:
            self.fail("jj_manhattan(design, \"my_name2\", options={})")

        try:
            jj_manhattan(design, "my_name3", options={}, make=False)
        except Exception:
            self.fail(
                "jj_manhattan(design, \"my_name3\", options={}, make=False)")


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from typing import Union, Tuple
import pandas as pd

from qiskit_metal.designs import QDesign  # pylint: disable=unused-import

from ... import Dict
from ..core import QSimulation


class ScatteringImpedanceSim(QSimulation):
    """Uses drivenmodal simulation to extract the impedance.

    Default Setup:
        * name (str): Name of driven modal setup. Defaults to "Setup".
        * freq_ghz (int): Frequency in GHz. Defaults to 5.
        * max_delta_s (float): Absolute value of maximum difference in
            scattering parameter S. Defaults to 0.1.
        * max_passes (int): Maximum number of passes. Defaults to 10.
        * min_passes (int): Minimum number of passes. Defaults to 1.
        * min_converged (int): Minimum number of converged passes. Defaults to 1.
        * pct_refinement (int): Percent refinement. Defaults to 30.
        * basis_order (int): Basis order. Defaults to 1.
        * vars (Dict): Variables (key) and values (value) to define in the renderer.
        * sweep_setup (Dict): Description of the drivenmodal sweep
            * name (str): Name of sweep. Defaults to "Sweep".
            * start_ghz (float): Starting frequency of sweep in GHz. Defaults to 2.0.
            * stop_ghz (float): Ending frequency of sweep in GHz. Defaults to 8.0.
            * count (int): Total number of frequencies. Defaults to 101.
            * step_ghz (float): Difference between adjacent frequencies. Defaults to None.
            * type (str): Type of sweep. Defaults to "Fast".
            * save_fields (bool): Whether or not to save fields. Defaults to False.

    Data Labels:
        * sweep_name (str): Name given to the current sweep.
        * param_z (pd.DataFrame): Impedance matrix.
        * param_y (pd.DataFrame): Admittance matrix.
        * param_s (pd.DataFrame): Scattering matrix.

    """
    default_setup = Dict(freq_ghz=5,
                         max_delta_s=0.1,
                         max_passes=10,
                         min_passes=1,
                         min_converged=1,
                         pct_refinement=30,
                         basis_order=1,
                         vars=Dict(Lj='10 nH', Cj='0 fF'),
                         sweep_setup=Dict(name="Sweep",
                                          start_ghz=2.0,
                                          stop_ghz=8.0,
                                          count=101,
                                          step_ghz=None,
                                          type="Fast",
                                          save_fields=False))
    """Default setup."""

    # supported labels for data generated from the simulation
    data_labels = ['sweep_name', 'params_z', 'params_y', 'params_s']
    """Default data labels."""

    def __init__(self, design: 'QDesign' = None, renderer_name: str = 'hfss'):
        """Compute drivenmodal and then extracts impedance, admittance and scattering paramters.

        Args:
            design (QDesign): Pointer to the main qiskit-metal design.
                Used to access the QRenderer. Defaults to None.
            renderer_name (str, optional): Which renderer to use. Defaults to 'hfss'.
        """
        # set design and renderer
        super().__init__(design, renderer_name)

    def _analyze(self):
        """Executes the analysis step of the Run. First it initializes the renderer setup
        to prepare for drivenmodal analysis, then it executes it. Finally it recovers the
        output of the analysis and stores it in self.params_z/params_y/params_s.
        """
        self.sim_setup_name, self.sweep_name = self.renderer.initialize_drivenmodal(
            **self.setup)

        self.renderer.analyze_sweep(self.sweep_name, self.sim_setup_name)
        # TODO: return the impedance, admittance and scattering matrices for later use

    def get_impedance(self, param_name: list = ['Z11', 'Z21']):
        """Create the impedance plot.

        Args:
            param_name (list): List of strings describing which impedance values
                to return. Defaults to ['Z11', 'Z21'].
        """
        # TODO: move the plot-making to this analysis module. Renderer should recover full data
        return self.renderer.plot_params(param_name)

    def get_admittance(self, param_name: list = ['Y11', 'Y21']):
        """Create the impedance plot.

        Args:
            param_name (list): List of strings describing which admittance values
                to return. Defaults to ['Y11', 'Y21'].
        """
        # TODO: move the plot in this analysis module. Renderer should recover the entire data
        return self.renderer.plot_params(param_name)

    def get_scattering(self, param_name: list = ['S11', 'S21']):
        """Create the scattering plot.

        Args:
            param_name (list): List of strings describing which scattering values
                to return. Defaults to ['S11', 'S21'].
        """
        # TODO: move the plot in this analysis module. Renderer should recover the entire data
        return self.renderer.plot_params(param_name)

    def run_sim(  # pylint: disable=arguments-differ
            self,
            name: str = None,
            components: Union[list, None] = None,
            open_terminations: Union[list, None] = None,
            port_list: Union[list, None] = None,
            jj_to_port: Union[list, None] = None,
            ignored_jjs: Union[list, None] = None,
            box_plus_buffer: bool = True) -> Tuple[str, str]:
        """Executes the entire drivenmodal analysis and convergence result export.
        First it makes sure the tool is running. Then it does what's necessary to render the design.
        Finally it runs the setup and sweep defined in this class. You need to modify the setup
        ahead. You can modify the setup by using the methods defined in the QAnalysis super-class.
        After this method concludes you can inspect the output using this class properties.

        Args:
            name (str): reference name for the components selection. If None,
                it will use the design.name. Defaults to None.
            components (Union[list, None], optional): List of components to render.
                Defaults to None.
            open_terminations (Union[list, None], optional):
                List of tuples of pins that are open. Defaults to None.
            port_list (Union[list, None], optional): List of tuples of pins to be rendered as ports.
                Format element: (component_name, pin_name, impedance (float)). Defaults to None.
            jj_to_port (Union[list, None], optional): List of tuples of jj's to be rendered as
                ports. Format element: (component_name(str), element_name(str), impedance(float),
                draw_ind(bool)). If draw_ind=True, a 10nH Inductance is draw, else it is omitted.
                Defaults to None.
            ignored_jjs (Union[list, None], optional): List of tuples of jj's that shouldn't be
                rendered. Defaults to None.
            box_plus_buffer (bool, optional): Either calculate a bounding box based on the location
                of rendered geometries or use chip size from design class. Defaults to True.

        Returns:
            (str, str): Name of the design and name of the setup
        """
        # save input variables to run(). This line must be the first in the method
        if components is not None:
            argm = locals()
            del argm['self']
            self.save_run_args(**argm)
        # wipe data from the previous run (if any)
        self.clear_data()

        if not self.renderer_initialized:
            self._initialize_renderer()

        vars_to_initialize = self.setup.vars
        renderer_design_name = self._render(
            name=name,
            solution_type='drivenmodal',
            selection=components,
            open_pins=open_terminations,
            port_list=port_list,
            jj_to_port=jj_to_port,
            ignored_jjs=ignored_jjs,
            box_plus_buffer=box_plus_buffer,
            vars_to_initialize=vars_to_initialize)

        self._analyze()
        return renderer_design_name, self.sim_setup_name

    @property
    def sweep_name(self) -> str:
        """Getter

        Returns:
            str: Name of the sweep being executed.
        """
        return self.get_data('sweep_name')

    @sweep_name.setter
    def sweep_name(self, data: str):
        """Setter

        Args:
            data (str): Name of the sweep being executed.
        """
        if not isinstance(data, str):
            self.logger.warning(
                'Unsupported type %s. Only accepts str. Please try again.',
                {type(data)})
            return
        self.set_data('sweep_name', data)

    @property
    def param_z(self) -> pd.DataFrame:
        """Getter

        Returns:
            str: Impedance matrix.
        """
        return self.get_data('param_z')

    @param_z.setter
    def param_z(self, data: pd.DataFrame):
        """Setter

        Args:
            data (pd.DataFrame): Impedance matrix.
        """
        if not isinstance(data, pd.DataFrame):
            self.logger.warning(
                'Unsupported type %s. Only accepts pd.DataFrame. Please try again.',
                {type(data)})
            return
        self.set_data('param_z', data)

    @property
    def param_y(self) -> pd.DataFrame:
        """Getter

        Returns:
            str: Admittance matrix.
        """
        return self.get_data('param_y')

    @param_y.setter
    def param_y(self, data: pd.DataFrame):
        """Setter

        Args:
            data (pd.DataFrame): Admittance matrix.
        """
        if not isinstance(data, pd.DataFrame):
            self.logger.warning(
                'Unsupported type %s. Only accepts pd.DataFrame. Please try again.',
                {type(data)})
            return
        self.set_data('param_y', data)

    @property
    def param_s(self) -> pd.DataFrame:
        """Getter

        Returns:
            str: Scattering matrix.
        """
        return self.get_data('param_s')

    @param_s.setter
    def param_s(self, data: pd.DataFrame):
        """Setter

        Args:
            data (pd.DataFrame): Scattering matrix.
        """
        if not isinstance(data, pd.DataFrame):
            self.logger.warning(
                'Unsupported type %s. Only accepts pd.DataFrame. Please try again.',
                {type(data)})
            return
        self.set_data('param_s', data)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
=================================================
QGeometryTables (:mod:`qiskit_metal.qgeometries`)
=================================================

.. currentmodule:: qiskit_metal.qgeometries

This modules defines and handles everything related to the qgeometry abstraction.

QGeometryTables
---------------

.. autosummary::
    :toctree: .

    QGeometryTables

"""
from .qgeometries_handler import is_qgeometry_table, QGeometryTables  # , QGeometry Types



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Module containing Multi-Planar design for CPW type
geometry. Supports tek file approach for layer stack definitions."""

#from typing import Tuple

from qiskit_metal.designs.design_base import QDesign
from qiskit_metal.toolbox_metal.layer_stack_handler import LayerStackHandler
from addict import Dict
from typing import Tuple

__all__ = ['MultiPlanar']


class MultiPlanar(QDesign):
    """Metal class for a multiple planar design, consisting of either single or multiple chips.
    Typically assumed to have some CPW geometries.

    Inherits QDesign class.
    """

    def __init__(self,
                 metadata: dict = None,
                 overwrite_enabled: bool = False,
                 enable_renderers: bool = True,
                 layer_stack_filename: str = None):
        """Pass metadata to QDesign.

        Args:
            metadata (dict, optional): Pass to QDesign. Defaults to {}.
            overwrite_enabled (bool, optional): Passed to QDesign base class. Defaults to False.
            enable_renderers (bool, optional): Passed to QDesign base class. Defaults to True.
        """

        super().__init__(metadata=metadata,
                         overwrite_enabled=overwrite_enabled,
                         enable_renderers=enable_renderers)

        #just using the single planar approach for the moment
        #still have the different chips, or just all via layers?
        self._add_chip_info()

        #generates the table for the layer information
        self.ls_filename = layer_stack_filename

        # Note: layers numbers can be repeated since there can be datatypes.
        self.ls = self._add_layer_stack()

        self._uwave_package = Dict()
        self._populate_uwave_values()

    def _populate_uwave_values(self):
        """Can be updated by user.
        """
        self._uwave_package[
            'sample_holder_top'] = '890um'  # how tall is the vacuum above center_z
        self._uwave_package[
            'sample_holder_bottom'] = '1650um'  # how tall is the vacuum below z=0

    def _add_layer_stack(self) -> LayerStackHandler:
        """Adds the data structure for the "layer_stack file" in the design for defining
        the layer stack. Simple initial layer is generated (to support the default
        layer used in all components).
        """
        return LayerStackHandler(self)

    def _add_chip_info(self):
        """Used to determine size of fill box by either 'size' data or box_plus_buffer.

        GDSTK is using numbers based on 1 meter unit.
        When the gds file is exported, data is converted to "user-selected" units.
        centered at (0,0) and 9 by 6 size.

        NOTE: self._chips dict comes from QDesign base class.
        """
        self._chips['main'] = Dict()

        self._chips['main']['size'] = Dict(
            center_x='0.0mm',
            center_y='0.0mm',
            size_x='9mm',
            size_y='7mm',
        )

    def get_x_y_for_chip(self, chip_name: str) -> Tuple[tuple, int]:
        """If the chip_name is in self.chips, along with entry for size
        information then return a tuple=(minx, miny, maxx, maxy). Used for
        subtraction while exporting design.

        Args:
            chip_name (str): Name of chip that you want the size of.

        Returns:
            Tuple[tuple, int]:
            tuple: The exact placement on rectangle coordinate (minx, miny, maxx, maxy).
            int: 0=all is good
            1=chip_name not in self._chips
            2=size information missing or no good
        """
        x_y_location = tuple()

        if chip_name in self._chips:
            if 'size' in self._chips[chip_name]:

                size = self.parse_value(self.chips[chip_name]['size'])
                if      'center_x' in size               \
                    and 'center_y' in size          \
                    and 'size_x' in size            \
                    and 'size_y' in size:
                    if type(size.center_x) in [int, float] and \
                            type(size.center_y) in [int, float] and \
                            type(size.size_x) in [int, float] and \
                            type(size.size_y) in [int, float]:
                        x_y_location = (
                            size['center_x'] - (size['size_x'] / 2.0),
                            size['center_y'] - (size['size_y'] / 2.0),
                            size['center_x'] + (size['size_x'] / 2.0),
                            size['center_y'] + (size['size_y'] / 2.0))
                        return x_y_location, 0

                    self.logger.warning(
                        f'Size information within self.chips[{chip_name}]["size"]'
                        f' is NOT an int or float.')
                    return x_y_location, 2

                self.logger.warning('center_x or center_y or size_x or size_y '
                                    f' NOT in self._chips[{chip_name}]["size"]')
                return x_y_location, 2

            self.logger.warning(
                f'Information for size in NOT in self._chips[{chip_name}]'
                ' dict. Return "None" in tuple.')
            return x_y_location, 2

        self.logger.warning(
            f'Chip name "{chip_name}" is not in self._chips dict. Return "None" in tuple.'
        )
        return x_y_location, 1



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Handles editing a QComponent."""

import ast
import inspect
from inspect import getfile, signature
from pathlib import Path
from typing import TYPE_CHECKING, Union
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtGui import QFont

__all__ = ['parse_param_from_str']

if TYPE_CHECKING:
    from .component_widget import ComponentWidget
    from ....designs import QDesign


class QTableModel_Options(QAbstractTableModel):
    """Table model for the options of a given component.

    This class inherits from the `QAbstractTableModel` class

    MVC class
    See https://doc.qt.io/qt-5/qabstracttablemodel.html
    """

    # __timer_interval = 500  # ms

    def __init__(self,
                 gui: 'MetalGUI',
                 parent: 'ComponentWidget' = None,
                 view=None):
        """
        Args:
            gui (MetalGUI): The GUI
            parent (ComponentWidget): The parent ComponentWidget.  Defaults to None.
            view (object): The view.  Defaults to None.
        """
        super().__init__(parent=parent)
        self.logger = gui.logger
        self.gui = gui
        self._row_count = -1
        self._view = view

        # self._create_timer()
        self.columns = ['Name', 'Value', 'Parsed value']

    @property
    def design(self) -> 'QDesign':
        """Returns the QDesign."""
        return self.gui.design

    @property
    def component(self):
        """Returns the component."""
        return self.parent().component

    def refresh(self):
        """Force refresh.

        Completly rebuild the model.
        """
        self.modelReset.emit()

    def rowCount(self, parent: QModelIndex = None):
        """Returns the number of rows.

        Args:
            parent (QModelIndex): Unused.  Defaults to None.

        Returns:
            int: The number of rows
        """
        if self.component is None:
            if self._view:
                self._view.show_placeholder_text()
            return 0
        if self._view:
            self._view.hide_placeholder_text()
        return len(self.component.options)

    def columnCount(self, parent: QModelIndex = None):
        """Returns the number of columns.

        Args:
            parent (QModelIndex): Unused.  Defaults to None.

        Returns:
            int: The number of columns
        """
        return 3

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Set the headers to be displayed.

        Args:
            section (int): Section number
            orientation (Qt orientation): Section orientation
            role (Qt display role): Display role.  Defaults to DisplayRole.

        Returns:
            str: The header data, or None if not found
        """

        if self.component is None:
            return None

        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if section < self.columnCount():
                    return self.columns[section]

        elif role == Qt.FontRole:
            if section == 0:
                font = QFont()
                font.setBold(True)
                return font

    def flags(self, index: QModelIndex):
        """Set the item flags at the given index. Seems like we're implementing
        this function just to see how it's done, as we manually adjust each
        treeView to have NoEditTriggers.

        Args:
            index (QModelIndex): The index

        Returns:
            Qt flags: Flags from Qt
        """
        # https://doc.qt.io/qt-5/qt.html#ItemFlag-enum

        if not index.isValid():
            return Qt.ItemIsEnabled

        # Returns the item flags for the given index.
        # The base class implementation returns a combination of flags that enables
        # the item (ItemIsEnabled) and allows it to be selected (ItemIsSelectable).
        flags = QAbstractTableModel.flags(self, index)
        if index.column() == 1:
            flags |= Qt.ItemIsEditable

        return Qt.ItemFlags(flags)

    # https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum
    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        """Depending on the index and role given, return data. If not returning
        data, return None (PySide equivalent of QT's "invalid QVariant").

        Returns:
            str: Data matching theindex and role.
        """

        if not index.isValid():
            return
        # if not 0 <= index.row() < self.rowCount():
        #    return None

        if self.component is None:
            return

        # The key data to be rendered in the form of text. (QString)
        if role == Qt.DisplayRole:
            row = index.row()
            column = index.column()
            data = self.component.options
            # There's probably a better way to access the data here
            if column == 0:
                data = list(data.keys())
            elif column in [1, 2]:
                data = list(data.values())

            data = data[row]
            if column == 2:
                if isinstance(data, dict):
                    return 'This is a dictionary.'
                else:
                    return str(self.design.parse_value(data))
            else:
                return str(data)

        # The data in a form suitable for editing in an editor.  (QString)
        elif role == Qt.EditRole:
            return self.data(index, QtCore.Qt.DisplayRole)

        # The font used for items rendered with the default delegate. (QFont)
        elif role == Qt.FontRole:
            if index.column() == 0:
                font = QtGui.QFont()
                font.setBold(True)
                return font

    def setData(self,
                index: QtCore.QModelIndex,
                value,
                role=QtCore.Qt.EditRole) -> bool:
        """Sets the role data for the item at index to value. The dataChanged()
        signal should be emitted if the data was successfully set.

        Args:
            index (QtCore.QModelIndex): The index
            value (str): The value
            role (QtCore.Qt.EditRole): The edit role

        Returns:
            bool: Returns true if successful; otherwise returns false.
        """

        # TODO: handle nested dictionaries
        # See v0.1: get_nested_dict_item, pop_nested_dict_item
        # TODO: ability to add dictionary such as to add pins
        if not index.isValid():
            return False

        elif role == QtCore.Qt.EditRole:
            if index.column() == 1:
                self._value = value  # QString
                value = str(value)

                data = self.component.options  # type: dict
                key, old_val = list(data.items())[index.row()]

                # When we do nothing
                if isinstance(old_val, dict):
                    self.logger.error(
                        'You selected a dicitonary this'
                        'cannot be edited directly edit its items.')
                    return False

                if old_val == value:
                    return False

                # When we do something to change the value

                # try:
                # TODO: should retry and if error then reset the value
                if 1:
                    self.logger.info(
                        f'Component options: Old value={old_val}; New value={value};'
                    )
                    if isinstance(old_val, str):
                        data[key] = str(value)
                    else:
                        processed_value, used_ast = parse_param_from_str(value)
                        self.logger.info(
                            f'  Used paring:  Old value type={type(old_val)}; '
                            f'New value type={type(processed_value)};  New value={processed_value};'
                            f'; Used ast={used_ast}')
                        data[key] = processed_value

                    self.component.rebuild()
                    self.gui.refresh()

                # except and finally restore the value
                return True

        # elif role == Qt.CheckStateRole:

        return False


def parse_param_from_str(text):
    """Attempt to parse a value from a string using ast.

    Args:
        text (str): String to parse

    Return:
        tuple: value, used_ast

    Raises:
        Exception: An error occurred
    """
    text = str(text).strip()
    value = text
    used_ast = False
    try:  # crude way to handle list and values
        value = ast.literal_eval(text)
        used_ast = True
    except Exception as exception:
        pass
        # print(exception)
    return value, used_ast



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=broad-except
# pylint: disable-msg=import-error
# pylint: disable-msg=too-many-public-methods
"""Qiskit Metal unit tests analyses functionality."""

from pathlib import Path
import unittest

import numpy as np
import pandas as pd

from qiskit_metal.analyses.quantization import lumped_capacitive
from qiskit_metal.analyses.hamiltonian.transmon_charge_basis import Hcpb
from qiskit_metal.analyses.hamiltonian.HO_wavefunctions import wavefunction
from qiskit_metal.analyses.em import cpw_calculations, kappa_calculation
from qiskit_metal.analyses.sweep_and_optimize.sweeper import Sweeper
from qiskit_metal.tests.assertions import AssertionsMixin
from qiskit_metal import designs

TEST_DATA = Path(__file__).parent / "test_data"


class TestAnalyses(unittest.TestCase, AssertionsMixin):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_analyses_cpw_guided_wavelength(self):
        """Test the functionality of guided_wavelength in
        cpw_calculations.py."""
        # Setup expected test results
        test_a_expected = (0.024345363624151843, 2.46280979514796,
                           0.49998185730956207)
        test_b_expected = (0.02900577391226639, 2.460840094957294,
                           0.4999524962937447)
        test_c_expected = (0.02135485699188455, 2.807698502630373,
                           0.49998185730956207)
        test_d_expected = (0.025443724808268103, 2.805350708925153,
                           0.4999524962937447)

        # Generate actual result data
        test_a_actual = cpw_calculations.guided_wavelength(
            5 * 10**9, 10 * 10**-6, 6 * 10**-6, 760 * 10**-6, 200 * 10**-9)
        test_b_actual = cpw_calculations.guided_wavelength(
            4.2 * 10**9, 9.8 * 10**-6, 4.8 * 10**-6, 420 * 10**-6, 180 * 10**-9)
        test_c_actual = cpw_calculations.guided_wavelength(
            5 * 10**9,
            10 * 10**-6,
            6 * 10**-6,
            760 * 10**-6,
            200 * 10**-9,
            dielectric_constant=15.2)
        test_d_actual = cpw_calculations.guided_wavelength(
            4.2 * 10**9,
            9.8 * 10**-6,
            4.8 * 10**-6,
            420 * 10**-6,
            180 * 10**-9,
            dielectric_constant=15.2)

        # Test all elements of the result data against expected data
        for (actual, expected) in [(test_a_actual, test_a_expected),
                                   (test_b_actual, test_b_expected),
                                   (test_c_actual, test_c_expected),
                                   (test_d_actual, test_d_expected)]:
            self.assertIterableAlmostEqual(actual, expected)

    def test_analyses_cpw_lumped_cpw(self):
        """Test the functionality of lumped_cpw in cpw_calculations.py."""
        # Setup expected test results
        expected = (0.04160856394637145, 3.7752144497164426e-07,
                    2.9513334836287544e-11, 0.0, 113.09975672923031,
                    1.0000000003871512, 2.9513334836287544e-11)

        # Generate actual result data
        result = cpw_calculations.lumped_cpw(1000000.75,
                                             0.25,
                                             0.1,
                                             0.002,
                                             0.003,
                                             dielectric_constant=15.1,
                                             loss_tangent=0.1,
                                             london_penetration_depth=1 *
                                             10**.7)

        # Test all elements of the result data against expected data
        self.assertIterableAlmostEqual(expected, result)

    def test_analyses_cpw_effective_dielectric_constant(self):
        """Test the functionality of effective_dielectric_constant in
        cpw_calculations.py."""
        # Generate actual result data
        test_a_actual = cpw_calculations.effective_dielectric_constant(
            5 * 10**9, 10 * 10**-6, 6 * 10**-6, 760 * 10**-6, 200 * 10**-9, 1.3,
            0.2, 1.3)
        test_b_actual = cpw_calculations.effective_dielectric_constant(
            4.8 * 10**9, 9.6 * 10**-6, 5.2 * 10**-6, 420 * 10**-6, 200 * 10**-9,
            1.3, 0.2, 1.3)
        test_c_actual = cpw_calculations.effective_dielectric_constant(
            4.8 * 10**9,
            9.6 * 10**-6,
            5.2 * 10**-6,
            420 * 10**-6,
            200 * 10**-9,
            1.3,
            0.2,
            1.3,
            eRD=200)

        # Test all elements of the result data against expected data
        for (actual, expected) in [(test_a_actual, 3.5771424204599436),
                                   (test_b_actual, 3.5442491595505388),
                                   (test_c_actual, 14.87179231388151)]:
            self.assertAlmostEqual(actual, expected, places=13)

        with self.assertRaises(ZeroDivisionError):
            test_a_actual = cpw_calculations.effective_dielectric_constant(
                5 * 10**9, 10 * 10**-6, 0, 760 * 10**-6, 200 * 10**-9, 1.3, 0.2,
                1.3)

        with self.assertRaises(ZeroDivisionError):
            test_a_actual = cpw_calculations.effective_dielectric_constant(
                5 * 10**9, 10 * 10**-6, 6 * 10**-6, 0, 200 * 10**-9, 1.3, 0.2,
                1.3)

    def test_analyses_cpw_elliptic_int_constants(self):
        """Test the functionality of elliptic_int_constants in
        cpw_calculations.py."""
        # Setup expected test results
        test_a_expected = (1.6629724332436984, 2.2424412123997364,
                           1.6629580628562461, 2.2425032046815248)
        test_b_expected = (1.6800926134149732, 2.1758110550759615,
                           1.6800526015534265, 2.175951807571837)

        # Generate actual result data
        test_a_actual = cpw_calculations.elliptic_int_constants(
            10 * 10**-6, 6 * 10**-6, 760 * 10**-6)
        test_b_actual = cpw_calculations.elliptic_int_constants(
            9.2 * 10**-6, 4.8 * 10**-6, 420 * 10**-6)

        # Test all elements of the result data against expected data
        for (actual, expected) in [(test_a_actual, test_a_expected),
                                   (test_b_actual, test_b_expected)]:
            self.assertIterableAlmostEqual(actual, expected)

        with self.assertRaises(ZeroDivisionError):
            cpw_calculations.elliptic_int_constants(0, 0, 0)

    def test_analysis_lumped_ic_from_lj(self):
        """Test the Ic_from_Lj function in lumped_capacitives.py."""
        self.assertAlmostEqualRel(lumped_capacitive.Ic_from_Lj(5e9),
                                  6.579465347418954e-26,
                                  rel_tol=1e-26)

    def test_analysis_lumped_ic_from_ej(self):
        """Test the Ic_from_Ej function in lumped_capacitives.py."""
        self.assertAlmostEqualRel(lumped_capacitive.Ic_from_Ej(5),
                                  1.5198803355538426e+16,
                                  rel_tol=1e16)

    def test_analyses_lumped_transmon_props(self):
        """Test the functionality of lumped_transmon_props in
        lumped_capacitives.py."""
        # Setup expected test results
        expected = [
            (3.2897326737094774e-12, 311949615351887.6, 0.00018137620223473302,
             1.2170673260260538, 55134023.405734204, 55134024.62280153, 0.0),
            (3.2897326737094773e-13, 3119496153518876.0, 1.81376202234733e-05,
             0.12170673260260537, 55134024.501094796, 55134024.62280153, 0.0),
            (3.289732673709477e-14, 3.1194961535188764e+16,
             1.8137620223473303e-06, 0.012170673260260537, 55134024.610630855,
             55134024.62280153, 0.0),
            (3.2897326737094773e-16, 3.119496153518876e+18,
             1.8137620223473303e-08, 0.00012170673260260537, 55134024.62267982,
             55134024.62280153, 0.0),
            (3.289732673709477e-17, 3.1194961535188763e+19,
             1.8137620223473302e-09, 1.2170673260260537e-05, 55134024.62278935,
             55134024.62280153, 0.0)
        ]

        # Generate actual result data
        result = []
        result.append(lumped_capacitive.transmon_props(0.0001, 0.0001))
        result.append(lumped_capacitive.transmon_props(0.001, 0.001))
        result.append(lumped_capacitive.transmon_props(0.01, 0.01))
        result.append(lumped_capacitive.transmon_props(1, 1))
        result.append(lumped_capacitive.transmon_props(10, 10))

        # Test all elements of the result data against expected data
        self.assertEqual(len(expected), len(result))
        for x, _ in enumerate(expected):
            self.assertAlmostEqual(_, result[x])

    def test_analyses_lumped_chi(self):
        """Test the functionality of chi in lumped_capacitives.py."""
        # Generate actual result data
        test_a_result = lumped_capacitive.chi(50, 30, 20, 10)
        test_b_result = lumped_capacitive.chi(100, 60, 40, 20)

        # Test all elements of the result data against expected data
        self.assertEqual(-137.5, test_a_result)
        self.assertEqual(-275.0, test_b_result)

        # need separate with's because if one works the with will bail
        with self.assertRaises(ZeroDivisionError):
            test_a_result = lumped_capacitive.chi(50, 30, 30, 10)

        with self.assertRaises(ZeroDivisionError):
            test_a_result = lumped_capacitive.chi(50, 30, 20, 30)

    def test_analyses_lumped_levels_vs_ng_real_units(self):
        """Test the functionality of levels_vs_ng_real_units in
        lumped_capacitives.py."""
        # Setup expected test results
        expected = [(388.69198590629856, 383542.9078924127, 774759511877.3142,
                     1.30777592387123e-06),
                    (776.6068552699635, 766320.2586084654, 1548045428680.8086,
                     6.545104023767585e-07),
                    (388691.64789067156, 383543414.14295447, 774809178857568.5,
                     1.3076920925450654e-09),
                    (402.9047503275388, 340910.07369285496, 774759511877.3142,
                     1.30777592387123e-06),
                    (804.9985387148533, 681147.9428256081, 1548045428680.8086,
                     6.545104023767585e-07),
                    (402900.7730357366, 340916038.71322423, 774809178857568.5,
                     1.3076920925450654e-09)]

        # Generate actual result data
        result = []
        result.append(lumped_capacitive.levels_vs_ng_real_units(0.1, 0.1))
        result.append(
            lumped_capacitive.levels_vs_ng_real_units(0.05005, 0.05005))
        result.append(lumped_capacitive.levels_vs_ng_real_units(0.0001, 0.0001))
        result.append(lumped_capacitive.levels_vs_ng_real_units(0.1, 0.1, N=25))
        result.append(
            lumped_capacitive.levels_vs_ng_real_units(0.05005, 0.05005, N=25))
        result.append(
            lumped_capacitive.levels_vs_ng_real_units(0.0001, 0.0001, N=25))

        # Test all elements of the result data against expected data
        for my_result in result:
            self.assertEqual(len(my_result), 4)

        self.assertEqual(len(expected), len(result))
        for x, _ in enumerate(expected):
            self.assertIterableAlmostEqual(_, result[x], rel_tol=1e-6)

        with self.assertRaises(ValueError):
            lumped_capacitive.levels_vs_ng_real_units(100, 100, N=-10)

    def test_analyses_lumped_get_c_and_ic(self):
        """Test the functionality of get_C_and_Ic in lumped_capacitives.py."""
        # Setup expected test results
        expected = [33.38125825, -0.61217795]

        # Generate actual result data
        result = lumped_capacitive.get_C_and_Ic(0.1, 0.2, 1.25, 1.75)

        # Test all elements of the result data against expected data
        self.assertEqual(len(result), 2)
        self.assertAlmostEqualRel(expected[0], result[0], rel_tol=1e-6)
        self.assertAlmostEqualRel(expected[1], result[1], rel_tol=1e-6)

    def test_analyses_lumped_cos_to_mega_and_delta(self):
        """Test the functionality of cos_to_mega_and_delta in
        lumped_capacitives.py."""
        # Generate actual result data
        test_a_result = lumped_capacitive.cos_to_mega_and_delta(
            1000, 100, 200, 200)
        test_b_result = lumped_capacitive.cos_to_mega_and_delta(
            1000, -100, 0.258, 200)

        # Test all elements of the result data against expected data
        self.assertAlmostEqual(278.9538182848156, test_a_result)
        self.assertAlmostEqual(197.27085868098266, test_b_result)

        with self.assertRaises(ZeroDivisionError):
            test_a_result = lumped_capacitive.cos_to_mega_and_delta(
                0, 100, 200, 200)

    def test_analyses_lumped_chargeline_t1(self):
        """Test the functionality of chargeline_T1 in lumped_capacitives.py."""
        # Generate actual result data
        test_a_result = lumped_capacitive.chargeline_T1(100, 20, 30)
        test_b_result = lumped_capacitive.chargeline_T1(13, 104, 1.6)

        # Test all elements of the result data against expected data
        self.assertAlmostEqual(1.1257909293593088e-09, test_a_result, places=13)
        self.assertAlmostEqual(0.00012178026880088676, test_b_result, places=13)

        with self.assertRaises(ZeroDivisionError):
            lumped_capacitive.chargeline_T1(1, 55, 0)

    def test_analyses_lumped_readin_q3d_matrix(self):
        """Test the functionality of readin_q3d_matrix in
        lumped_capacitives.py."""
        # Setup expected test results
        units_expected = 'farad'
        design_variation_expected = (
            "$BBoxL='650um' $boxH='750um' $boxL='2mm' $QubitGap='30um'" +
            " $QubitH='90um' $QubitL='450um' Lj_1='13nH'")

        data = {
            'ground_plane': [
                2.882900e-13, -3.254000e-14, -3.197800e-14, -4.006300e-14,
                -4.384200e-14, -3.005300e-14
            ],
            'Q1_bus_Q0_connector_pad': [
                -3.254000e-14, 4.725700e-14, -2.276500e-16, -1.269000e-14,
                -1.335100e-15, -1.451000e-16
            ],
            'Q1_bus_Q2_connector_pad': [
                -3.197800e-14, -2.276500e-16, 4.532700e-14, -1.218000e-15,
                -1.155200e-14, -5.041400e-17
            ],
            'Q1_pad_bot': [
                -4.006300e-14, -1.269000e-14, -1.218000e-15, 9.583100e-14,
                -3.241500e-14, -8.366500e-15
            ],
            'Q1_pad_top1': [
                -4.384200e-14, -1.335100e-15, -1.155200e-14, -3.241500e-14,
                9.132000e-14, -1.019900e-15
            ],
            'Q1_readout_connector_pad': [
                -3.005300e-14, -1.451000e-16, -5.041400e-17, -8.366500e-15,
                -1.019900e-15, 3.988400e-14
            ]
        }
        df_cmat_expected = pd.DataFrame(data,
                                        index=[
                                            'ground_plane',
                                            'Q1_bus_Q0_connector_pad',
                                            'Q1_bus_Q2_connector_pad',
                                            'Q1_pad_bot', 'Q1_pad_top1',
                                            'Q1_readout_connector_pad'
                                        ])

        data = {
            'ground_plane': [0, 0, 0, 0, 0, 0],
            'Q1_bus_Q0_connector_pad': [0, 0, 0, 0, 0, 0],
            'Q1_bus_Q2_connector_pad': [0, 0, 0, 0, 0, 0],
            'Q1_pad_bot': [0, 0, 0, 0, 0, 0],
            'Q1_pad_top1': [0, 0, 0, 0, 0, 0],
            'Q1_readout_connector_pad': [0, 0, 0, 0, 0, 0]
        }
        df_cond_expected = pd.DataFrame(data,
                                        index=[
                                            'ground_plane',
                                            'Q1_bus_Q0_connector_pad',
                                            'Q1_bus_Q2_connector_pad',
                                            'Q1_pad_bot', 'Q1_pad_top1',
                                            'Q1_readout_connector_pad'
                                        ])

        # Generate actual result data
        generated = lumped_capacitive.readin_q3d_matrix(TEST_DATA /
                                                        'q3d_example.txt')
        return_size_result = len(generated)
        df_cmat_result = generated[0]
        units_result = generated[1]
        design_variation_result = generated[2]
        df_cond_result = generated[3]

        # Test all elements of the result data against expected data
        self.assertEqual(4, return_size_result)
        self.assertEqual(units_expected, units_result)
        self.assertEqual(design_variation_expected, design_variation_result)

        for j in [
                'ground_plane', 'Q1_bus_Q0_connector_pad',
                'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                'Q1_readout_connector_pad'
        ]:
            self.assertIterableAlmostEqual(df_cmat_expected[j],
                                           df_cmat_result[j])

        for j in [
                'ground_plane', 'Q1_bus_Q0_connector_pad',
                'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                'Q1_readout_connector_pad'
        ]:
            self.assertIterableAlmostEqual(df_cond_expected[j],
                                           df_cond_result[j])

    #pylint: disable-msg=too-many-locals
    def test_analyses_lumped_load_q3d_capacitance_matrix(self):
        """Test the functionality of load_q3d_capacitance_matrix in
        lumped_capacitives.py."""
        # Setup expected test results
        test_a_design_variation_expected = (
            "$BBoxL='650um' $boxH='750um' $boxL='2mm' " +
            "$QubitGap='30um' $QubitH='90um' $QubitL='450um' " + "Lj_1='13nH'")

        test_b_design_variation_expected = (
            "$BBoxL='650um' $boxH='750um' $boxL='2mm' " +
            "$QubitGap='30um' $QubitH='90um' $QubitL='450um' " + "Lj_1='13nH'")

        data = {
            'ground_plane': [
                288.290, -32.540, -31.978, -40.063, -43.842, -30.053
            ],
            'Q1_bus_Q0_connector_pad': [
                -32.54000, 47.25700, -0.22765, -12.69000, -1.33510, -0.14510
            ],
            'Q1_bus_Q2_connector_pad': [
                -31.978000, -0.227650, 45.327000, -1.218000, -11.552000,
                -0.050414
            ],
            'Q1_pad_bot': [
                -40.0630, -12.6900, -1.2180, 95.8310, -32.4150, -8.3665
            ],
            'Q1_pad_top1': [
                -43.8420, -1.3351, -11.5520, -32.4150, 91.3200, -1.0199
            ],
            'Q1_readout_connector_pad': [
                -30.053000, -0.145100, -0.050414, -8.366500, -1.019900,
                39.884000
            ]
        }
        test_a_df_cmat_expected = pd.DataFrame(data,
                                               index=[
                                                   'ground_plane',
                                                   'Q1_bus_Q0_connector_pad',
                                                   'Q1_bus_Q2_connector_pad',
                                                   'Q1_pad_bot', 'Q1_pad_top1',
                                                   'Q1_readout_connector_pad'
                                               ])

        data = {
            'ground_plane': [0, 0, 0, 0, 0, 0],
            'Q1_bus_Q0_connector_pad': [0, 0, 0, 0, 0, 0],
            'Q1_bus_Q2_connector_pad': [0, 0, 0, 0, 0, 0],
            'Q1_pad_bot': [0, 0, 0, 0, 0, 0],
            'Q1_pad_top1': [0, 0, 0, 0, 0, 0],
            'Q1_readout_connector_pad': [0, 0, 0, 0, 0, 0]
        }
        test_a_df_cond_expected = pd.DataFrame(data,
                                               index=[
                                                   'ground_plane',
                                                   'Q1_bus_Q0_connector_pad',
                                                   'Q1_bus_Q2_connector_pad',
                                                   'Q1_pad_bot', 'Q1_pad_top1',
                                                   'Q1_readout_connector_pad'
                                               ])

        data = {
            'ground_plane': [
                0.288290, -0.032540, -0.031978, -0.040063, -0.043842, -0.030053
            ],
            'Q1_bus_Q0_connector_pad': [
                -0.032540, 0.047257, -0.000228, -0.012690, -0.001335, -0.000145
            ],
            'Q1_bus_Q2_connector_pad': [
                -0.031978, -0.000228, 0.045327, -0.001218, -0.011552, -0.000050
            ],
            'Q1_pad_bot': [
                -0.040063, -0.012690, -0.001218, 0.095831, -0.032415, -0.008367
            ],
            'Q1_pad_top1': [
                -0.043842, -0.001335, -0.011552, -0.032415, 0.091320, -0.001020
            ],
            'Q1_readout_connector_pad': [
                -0.030053, -0.000145, -0.000050, -0.008367, -0.001020, 0.039884
            ]
        }
        test_b_df_cmat_expected = pd.DataFrame(data,
                                               index=[
                                                   'ground_plane',
                                                   'Q1_bus_Q0_connector_pad',
                                                   'Q1_bus_Q2_connector_pad',
                                                   'Q1_pad_bot', 'Q1_pad_top1',
                                                   'Q1_readout_connector_pad'
                                               ])

        data = {
            'ground_plane': [0, 0, 0, 0, 0, 0],
            'Q1_bus_Q0_connector_pad': [0, 0, 0, 0, 0, 0],
            'Q1_bus_Q2_connector_pad': [0, 0, 0, 0, 0, 0],
            'Q1_pad_bot': [0, 0, 0, 0, 0, 0],
            'Q1_pad_top1': [0, 0, 0, 0, 0, 0],
            'Q1_readout_connector_pad': [0, 0, 0, 0, 0, 0]
        }
        test_b_df_cond_expected = pd.DataFrame(data,
                                               index=[
                                                   'ground_plane',
                                                   'Q1_bus_Q0_connector_pad',
                                                   'Q1_bus_Q2_connector_pad',
                                                   'Q1_pad_bot', 'Q1_pad_top1',
                                                   'Q1_readout_connector_pad'
                                               ])

        # Generate actual result data
        test_a_result = lumped_capacitive.load_q3d_capacitance_matrix(
            TEST_DATA / 'q3d_example.txt', _disp=False)
        test_b_result = lumped_capacitive.load_q3d_capacitance_matrix(
            TEST_DATA / 'q3d_example.txt', user_units='pF', _disp=False)

        test_a_return_size_result = len(test_a_result)
        test_a_df_cmat_result = test_a_result[0]
        test_a_user_units_result = test_a_result[1]
        test_a_design_variation_result = test_a_result[2]
        test_a_df_cond_result = test_a_result[3]

        test_b_return_size_result = len(test_b_result)
        test_b_df_cmat_result = test_b_result[0]
        test_b_user_units_result = test_b_result[1]
        test_b_design_variation_result = test_b_result[2]
        test_b_df_cond_result = test_b_result[3]

        # Test all elements of the result data against expected data
        self.assertEqual(4, test_a_return_size_result)
        self.assertEqual('fF', test_a_user_units_result)
        self.assertEqual(test_a_design_variation_expected,
                         test_a_design_variation_result)

        self.assertEqual(4, test_b_return_size_result)
        self.assertEqual('pF', test_b_user_units_result)
        self.assertEqual(test_b_design_variation_expected,
                         test_b_design_variation_result)

        data_points = test_a_df_cmat_expected['ground_plane'].size
        for i in range(data_points):
            for j in [
                    'ground_plane', 'Q1_bus_Q0_connector_pad',
                    'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                    'Q1_readout_connector_pad'
            ]:
                self.assertAlmostEqual(test_a_df_cmat_expected[j][i],
                                       test_a_df_cmat_result[j][i],
                                       places=3)

        data_points = test_a_df_cond_expected['ground_plane'].size
        for i in range(data_points):
            for j in [
                    'ground_plane', 'Q1_bus_Q0_connector_pad',
                    'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                    'Q1_readout_connector_pad'
            ]:
                self.assertAlmostEqual(test_a_df_cond_expected[j][i],
                                       test_a_df_cond_result[j][i],
                                       places=3)

        data_points = test_b_df_cmat_expected['ground_plane'].size
        for i in range(data_points):
            for j in [
                    'ground_plane', 'Q1_bus_Q0_connector_pad',
                    'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                    'Q1_readout_connector_pad'
            ]:
                self.assertAlmostEqual(test_b_df_cmat_expected[j][i],
                                       test_b_df_cmat_result[j][i],
                                       places=3)

        data_points = test_b_df_cond_expected['ground_plane'].size
        for i in range(data_points):
            for j in [
                    'ground_plane', 'Q1_bus_Q0_connector_pad',
                    'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                    'Q1_readout_connector_pad'
            ]:
                self.assertAlmostEqual(test_b_df_cond_expected[j][i],
                                       test_b_df_cond_result[j][i],
                                       places=3)

    def test_analyses_lumped_move_index_to(self):
        """Test the functionality of move_index_to in lumped_capacitives.py."""
        # Setup expected test results
        test_a_expected = [0, 2, 3, 1, 4]
        test_b_expected = [1, 2, 3, 4, 0]
        test_c_expected = [0, 3, 1, 2, 4]

        # Generate actual result data
        test_a_result = lumped_capacitive.move_index_to(1, 3, 5)
        test_b_result = lumped_capacitive.move_index_to(0, 4, 5)
        test_c_result = lumped_capacitive.move_index_to(3, 1, 5)

        # Test all elements of the result data against expected data
        self.assertTrue(np.array_equal(test_a_expected, test_a_result))
        self.assertTrue(np.array_equal(test_b_expected, test_b_result))
        self.assertTrue(np.array_equal(test_c_expected, test_c_result))

        with self.assertRaises(IndexError):
            lumped_capacitive.move_index_to(3, 1, -5)

        with self.assertRaises(TypeError):
            lumped_capacitive.move_index_to(3, 1.5, 5)

    def test_analyses_lumped_df_reorder_matrix_basis(self):
        """Test the functionality of df_reorder_matrix_basis in
        lumped_capacitives.py."""
        # Setup expected test results
        data = {
            'ground_plane': [
                2.8829E-13, -3.1978E-14, -4.0063E-14, -3.254E-14, -4.3842E-14,
                -3.0053E-14
            ],
            'Q1_bus_Q2_connector_pad': [
                -3.1978E-14, 4.5327E-14, -1.218E-15, -2.2765E-16, -1.1552E-14,
                -5.0414E-17
            ],
            'Q1_pad_bot': [
                -4.0063E-14, -1.218E-15, 9.5831E-14, -1.269E-14, -3.2415E-14,
                -8.3665E-15
            ],
            'Q1_bus_Q0_connector_pad': [
                -3.254E-14, -2.2765E-16, -1.269E-14, 4.7257E-14, -1.3351E-15,
                -1.451E-16
            ],
            'Q1_pad_top1': [
                -4.3842E-14, -1.1552E-14, -3.2415E-14, -1.3351E-15, 9.132E-14,
                -1.0199E-15
            ],
            'Q1_readout_connector_pad': [
                -3.0053E-14, -5.0414E-17, -8.3665E-15, -1.451E-16, -1.0199E-15,
                3.9884E-14
            ]
        }

        test_a_expected = pd.DataFrame(
            data,
            index=[
                'ground_plane', 'Q1_bus_Q2_connector_pad', 'Q1_pad_bot',
                'Q1_bus_Q0_connector_pad', 'Q1_pad_top1',
                'Q1_readout_connector_pad'
            ])

        data = {
            'Q1_pad_top1': [
                9.132E-14, -4.3842E-14, -1.3351E-15, -1.1552E-14, -3.2415E-14,
                -1.0199E-15
            ],
            'ground_plane': [
                -4.3842E-14, 2.8829E-13, -3.254E-14, -3.1978E-14, -4.0063E-14,
                -3.0053E-14
            ],
            'Q1_bus_Q0_connector_pad': [
                -1.3351E-15, -3.254E-14, 4.7257E-14, -2.2765E-16, -1.269E-14,
                -1.451E-16
            ],
            'Q1_bus_Q2_connector_pad': [
                -1.1552E-14, -3.1978E-14, -2.2765E-16, 4.5327E-14, -1.218E-15,
                -5.0414E-17
            ],
            'Q1_pad_bot': [
                -3.2415E-14, -4.0063E-14, -1.269E-14, -1.218E-15, 9.5831E-14,
                -8.3665E-15
            ],
            'Q1_readout_connector_pad': [
                -1.0199E-15, -3.0053E-14, -1.451E-16, -5.0414E-17, -8.3665E-15,
                3.9884E-14
            ]
        }

        test_b_expected = pd.DataFrame(data,
                                       index=[
                                           'Q1_pad_top1', 'ground_plane',
                                           'Q1_bus_Q0_connector_pad',
                                           'Q1_bus_Q2_connector_pad',
                                           'Q1_pad_bot',
                                           'Q1_readout_connector_pad'
                                       ])

        # Generate actual result data
        df_a = lumped_capacitive.readin_q3d_matrix(TEST_DATA /
                                                   'q3d_example.txt')[0]
        df_b = lumped_capacitive.readin_q3d_matrix(TEST_DATA /
                                                   'q3d_example.txt')[0]

        test_a_result = lumped_capacitive.df_reorder_matrix_basis(df_a, 1, 3)
        test_b_result = lumped_capacitive.df_reorder_matrix_basis(df_b, 4, 0)

        # Test all elements of the result data against expected data
        for j in [
                'ground_plane', 'Q1_bus_Q2_connector_pad', 'Q1_pad_bot',
                'Q1_bus_Q0_connector_pad', 'Q1_pad_top1',
                'Q1_readout_connector_pad'
        ]:
            self.assertIterableAlmostEqual(test_a_expected[j],
                                           test_a_result[j],
                                           abs_tol=1e-20)

        for j in [
                'ground_plane', 'Q1_bus_Q0_connector_pad',
                'Q1_bus_Q2_connector_pad', 'Q1_pad_bot', 'Q1_pad_top1',
                'Q1_readout_connector_pad'
        ]:
            self.assertIterableAlmostEqual(test_b_expected[j],
                                           test_b_result[j],
                                           abs_tol=1e-20)

        with self.assertRaises(IndexError):
            test_a_result = lumped_capacitive.df_reorder_matrix_basis(
                df_a, 1, 35)

    def test_analyses_hamiltonian_ho_wavefunction(self):
        """Test the wavefunction function in the HO_waefunction.py file."""
        x_range = np.linspace(-5, 5, 5)
        actual = wavefunction(1.0, 1.0, 0, x_range)

        expected = [
            2.10255658e-06, 2.47888124e-02, 5.64189584e-01, 2.47888124e-02,
            2.10255658e-06
        ]

        self.assertEqual(len(actual), len(expected))

        for x, _ in enumerate(actual):
            self.assertAlmostEqualRel(_, expected[x], rel_tol=1e-6)

    def test_analysis_transmon_charge_basis_evaluek(self):
        """Test the evaluek function in the Hcpb class."""
        hcpb = Hcpb(nlevels=15, Ej=13971.3, Ec=295.2, ng=0.001)
        self.assertAlmostEqualRel(hcpb.evalue_k(0),
                                  -11175.114908531536,
                                  rel_tol=1e-6)
        self.assertAlmostEqualRel(hcpb.evalue_k(2),
                                  -653.7652579628739,
                                  rel_tol=1e-6)

    def test_analysis_transmon_charge_basis_evec_k(self):
        """Test the evec_k function in the Hcpb class."""
        hcpb = Hcpb(nlevels=2, Ej=13971.3, Ec=295.2, ng=0.001)

        expected = [
            -0.434102035, -0.558197591, -0.000417109677, 0.557943545,
            0.434361255
        ]
        actual = hcpb.evec_k(1)
        if actual[0] * expected[0] < 0:
            actual *= -1

        self.assertIterableAlmostEqual(expected, actual, abs_tol=1e-4)

    def test_analysis_transmon_charge_basis_fij(self):
        """Test the fij function in the Hcpb class."""
        hcpb = Hcpb(nlevels=15, Ej=13971.3, Ec=295.2, ng=0.001)

        self.assertAlmostEqual(hcpb.fij(1, 2), 5090.160741580386)

    def test_analysis_transmon_charge_basis_anharm(self):
        """Test the anharm function in the Hcpb class."""
        hcpb = Hcpb(nlevels=15, Ej=13971.3, Ec=295.2, ng=0.001)
        self.assertAlmostEqual(hcpb.anharm(), -341.0281674078906)

    def test_analysis_transmon_charge_basis_n_ij(self):
        """Test the n_ij function in the Hcpb class."""
        hcpb = Hcpb(nlevels=15, Ej=13971.3, Ec=295.2, ng=0.001)
        self.assertAlmostEqual(hcpb.n_ij(1, 2), 1.4670047579229986)

    def test_analysis_kappa_calculation_kappa_in(self):
        """Test the kappa_in function in kappa_calculation.py."""
        self.assertAlmostEqual(
            kappa_calculation.kappa_in(5.0E9, 30.0E-15, 4.5E9),
            161144.37988054403)

    def test_analysis_sweeper_option_value(self):
        """Test the option_value function in the Sweeper class"""
        from abc import ABC
        from qiskit_metal.analyses.core.base import QAnalysis
        sweeper = Sweeper(QAnalysis)

        in_dict = {'a': 1, 'b': 'bee'}
        self.assertEqual(sweeper.option_value(in_dict, 'a'), 1)
        self.assertEqual(sweeper.option_value(in_dict, 'b'), 'bee')


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent
import numpy as np


class CapNInterdigitalTee(QComponent):
    """Generates a three pin (+) structure comprised of a primary two pin CPW
    transmission line, and a secondary one pin neighboring CPW transmission
    line that is capacitively coupled to the primary. Such a structure can be
    used, as an example, for generating CPW resonator hangars off of a
    transmission line. (0,0) represents the center position of the component.
    Setting finger length to 0 gives a simple gap capacitor.

    Inherits QComponent class.

    ::

                  (0,0)
        +--------------------------+
                    |
               --|-----|--
              |  |  |  |  |
              |-----|-----|
                    |
                    |
                    |
                    |
                    +

    .. image::
        CapNInterdigitalTee.png

    .. meta::
        :description: Cap N Interdigital Tee

    Options:
        * prime_width: '10um' -- The width of the trace of the two pin CPW transmission line
        * prime_gap: '6um' -- The dielectric gap of the two pin CPW transmission line
        * second_width: '10um' -- The width of the trace of the one pin CPW transmission line
        * second_gap: '6um' -- The dielectric gap of the one pin CPW transmission line (also for the capacitor)
        * cap_gap: '6um' -- The width of dielectric for the capacitive coupling (default same as second_gap)
        * cap_width: '10um' -- The width of the finger capacitor (default same as second)
        * finger_length: '20um' -- The depth of the charge islands of the capacitor
        * finger_count: '5' -- Number of fingers in the capacitor
        * cap_distance: '50um' -- Distance of capacitor from center transmission line
    """
    component_metadata = Dict(short_name='cpw',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_path='True')
    """Component metadata"""

    #Currently setting the primary CPW length based on the coupling_length
    #May want it to be it's own value that the user can control?
    default_options = Dict(prime_width='10um',
                           prime_gap='6um',
                           second_width='10um',
                           second_gap='6um',
                           cap_gap='6um',
                           cap_width='10um',
                           finger_length='20um',
                           finger_count='5',
                           cap_distance='50um')
    """Default connector options"""

    TOOLTIP = """Generates a three pin (+) structure comprised of a primary two pin CPW
    transmission line, and a secondary one pin neighboring CPW transmission
    line that is capacitively coupled to the primary."""

    def make(self):
        """Build the component."""
        p = self.p
        N = int(p.finger_count)
        prime_cpw_length = p.cap_width * 2 * N

        #Primary CPW
        prime_cpw = draw.LineString([[-prime_cpw_length / 2, 0],
                                     [prime_cpw_length / 2, 0]])

        #Finger Capacitor
        cap_box = draw.rectangle(N * p.cap_width + (N - 1) * p.cap_gap,
                                 p.cap_gap + 2 * p.cap_width + p.finger_length,
                                 0, 0)
        make_cut_list = []
        make_cut_list.append([0, (p.finger_length) / 2])
        make_cut_list.append([(p.cap_width) + (p.cap_gap / 2),
                              (p.finger_length) / 2])
        flip = -1

        for i in range(1, N):
            make_cut_list.append([
                i * (p.cap_width) + (2 * i - 1) * (p.cap_gap / 2),
                flip * (p.finger_length) / 2
            ])
            make_cut_list.append([
                (i + 1) * (p.cap_width) + (2 * i + 1) * (p.cap_gap / 2),
                flip * (p.finger_length) / 2
            ])
            flip = flip * -1

        cap_cut = draw.LineString(make_cut_list).buffer(p.cap_gap / 2,
                                                        cap_style=2,
                                                        join_style=2)
        cap_cut = draw.translate(cap_cut,
                                 -(N * p.cap_width + (N - 1) * p.cap_gap) / 2,
                                 0)

        cap_body = draw.subtract(cap_box, cap_cut)
        cap_body = draw.translate(
            cap_body, 0, -p.cap_distance -
            (p.cap_gap + 2 * p.cap_width + p.finger_length) / 2)

        cap_etch = draw.rectangle(
            N * p.cap_width + (N - 1) * p.cap_gap + 2 * p.second_gap,
            p.cap_gap + 2 * p.cap_width + p.finger_length + 2 * p.second_gap, 0,
            -p.cap_distance -
            (p.cap_gap + 2 * p.cap_width + p.finger_length) / 2)

        #Secondary CPW
        second_cpw_top = draw.LineString([[0, -p.prime_width / 2],
                                          [0, -p.cap_distance]])

        second_cpw_bottom = draw.LineString(
            [[
                0, -p.cap_distance -
                (p.cap_gap + 2 * p.cap_width + p.finger_length)
            ],
             [
                 0, -2 * p.cap_distance -
                 (p.cap_gap + 2 * p.cap_width + p.finger_length)
             ]])

        #Rotate and Translate
        c_items = [
            prime_cpw, second_cpw_top, second_cpw_bottom, cap_body, cap_etch
        ]
        c_items = draw.rotate(c_items, p.orientation, origin=(0, 0))
        c_items = draw.translate(c_items, p.pos_x, p.pos_y)
        [prime_cpw, second_cpw_top, second_cpw_bottom, cap_body,
         cap_etch] = c_items

        #Add to qgeometry tables
        self.add_qgeometry('path', {'prime_cpw': prime_cpw},
                           width=p.prime_width,
                           layer=p.layer)
        self.add_qgeometry('path', {'prime_cpw_sub': prime_cpw},
                           width=p.prime_width + 2 * p.prime_gap,
                           subtract=True,
                           layer=p.layer)
        self.add_qgeometry('path', {
            'second_cpw_top': second_cpw_top,
            'second_cpw_bottom': second_cpw_bottom
        },
                           width=p.second_width,
                           layer=p.layer)
        self.add_qgeometry('path', {
            'second_cpw_top_sub': second_cpw_top,
            'second_cpw_bottom_sub': second_cpw_bottom
        },
                           width=p.second_width + 2 * p.second_gap,
                           subtract=True,
                           layer=p.layer)

        self.add_qgeometry('poly', {'cap_body': cap_body}, layer=p.layer)
        self.add_qgeometry('poly', {'cap_etch': cap_etch},
                           layer=p.layer,
                           subtract=True)

        #Add pins
        prime_pin_list = prime_cpw.coords
        second_pin_list = second_cpw_bottom.coords

        self.add_pin('prime_start',
                     points=np.array(prime_pin_list[::-1]),
                     width=p.prime_width,
                     input_as_norm=True)
        self.add_pin('prime_end',
                     points=np.array(prime_pin_list),
                     width=p.prime_width,
                     input_as_norm=True)
        self.add_pin('second_end',
                     points=np.array(second_pin_list),
                     width=p.second_width,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6 import QtCore
from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItem, QStandardItemModel


class DynamicList(QStandardItemModel):
    """Create a dynamic checklist that gets updated when the data source is
    modified or a new one is introduced."""

    def __init__(self, orig_design: 'QDesign'):
        """Set the original design source and populate list.

        Args:
            orig_design (QDesign): The design
        """
        super().__init__()
        self._design = orig_design
        self.populate_list()

    @property
    def datasrc(self):
        """Return the data source."""
        return self._design

    def update_src(self, new_design: 'QDesign'):
        """Change the data source to a new one.

        Args:
            new_design (QDesign): The new data source
        """
        self._design = new_design

    def populate_list(self):
        """Clear model and (re)populate it with the latest elements."""
        # TODO: Generalize this to beyond components.
        self.clear()
        if self.datasrc:
            for element in self.datasrc.components:
                item = QStandardItem(element)
                item.setCheckable(True)
                item.setCheckState(QtCore.Qt.Checked)
                item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
                self.appendRow(item)

    def select_all(self):
        """Select everything in the list."""
        for i in range(self.rowCount()):
            self.item(i).setCheckState(QtCore.Qt.Checked)

    def deselect_all(self):
        """Deselect everything in the list."""
        for i in range(self.rowCount()):
            self.item(i).setCheckState(QtCore.Qt.Unchecked)

    def get_checked(self) -> list:
        """Get list of all selected items.

        Returns:
            list: List of selected (checked) items
        """
        selected_items = []
        for i in range(self.rowCount()):
            entry = self.item(i)
            if entry.checkState() == QtCore.Qt.Checked:
                selected_items.append(entry.text())
        return selected_items



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent
import numpy as np


class CoupledLineTee(QComponent):
    """Generates a three pin (+) structure comprised of a primary two pin CPW
    transmission line, and a secondary one pin neighboring CPW transmission
    line that is capacitively/inductively coupled to the primary. Such a
    structure can be used, as an example, for generating CPW resonator hangars
    off of a transmission line.

    Inherits QComponent class.

    ::

        +----------------------------+
        ------------------------------
        |
        |
        |
        |
        +

    .. image::
        CoupledLineTee.png

    .. meta::
        :description: Coupled Line Tee

    Default Options:
        * prime_width: '10um' -- The width of the trace of the two pin CPW transmission line
        * prime_gap: '6um' -- The dielectric gap of the two pin CPW transmission line
        * second_width: '10um' -- The width of the trace of the one pin CPW transmission line
        * second_gap: '6um' -- The dielectric gap of the one pin CPW transmission line
        * coupling_space: '3um' -- The amount of ground plane between the two transmission lines
        * coupling_length: '100um' -- The length of parallel between the two transmission lines
          note: this includes the distance of the curved second of the second line
        * down_length: '100um' -- The length of the hanging part of the resonator, including the curved region
        * fillet: '25um'
        * mirror: False -- Flips the hanger around the y-axis
        * open_termination: True -- sets if the termination of the second line at the coupling side
          is an open to ground or short to ground
    """
    component_metadata = Dict(short_name='cpw', _qgeometry_table_path='True')
    """Component metadata"""

    #Currently setting the primary CPW length based on the coupling_length
    #May want it to be it's own value that the user can control?
    default_options = Dict(prime_width='10um',
                           prime_gap='6um',
                           second_width='10um',
                           second_gap='6um',
                           coupling_space='3um',
                           coupling_length='100um',
                           down_length='100um',
                           fillet='25um',
                           mirror=False,
                           open_termination=True)
    """Default connector options"""

    TOOLTIP = """Generates a three pin (+) 
    structure comprised of a primary two 
    pin CPW transmission line, and a 
    secondary one pin neighboring CPW 
    transmission line that is 
    capacitively/inductively coupled 
    to the primary."""

    def make(self):
        """Build the component."""
        p = self.p

        prime_cpw_length = p.coupling_length * 2
        second_flip = 1
        if p.mirror:
            second_flip = -1

        #Primary CPW
        prime_cpw = draw.LineString([[-prime_cpw_length / 2, 0],
                                     [prime_cpw_length / 2, 0]])

        #Secondary CPW
        second_down_length = p.down_length
        second_y = -p.prime_width / 2 - p.prime_gap - p.coupling_space - p.second_gap - p.second_width / 2
        second_cpw = draw.LineString(
            [[second_flip * (-p.coupling_length / 2), second_y],
             [second_flip * (p.coupling_length / 2), second_y],
             [
                 second_flip * (p.coupling_length / 2),
                 second_y - second_down_length
             ]])

        second_termination = 0
        if p.open_termination:
            second_termination = p.second_gap

        second_cpw_etch = draw.LineString(
            [[
                second_flip * (-p.coupling_length / 2 - second_termination),
                second_y
            ], [second_flip * (p.coupling_length / 2), second_y],
             [
                 second_flip * (p.coupling_length / 2),
                 second_y - second_down_length
             ]])

        #Rotate and Translate
        c_items = [prime_cpw, second_cpw, second_cpw_etch]
        c_items = draw.rotate(c_items, p.orientation, origin=(0, 0))
        c_items = draw.translate(c_items, p.pos_x, p.pos_y)
        [prime_cpw, second_cpw, second_cpw_etch] = c_items

        #Add to qgeometry tables
        self.add_qgeometry('path', {'prime_cpw': prime_cpw},
                           width=p.prime_width)
        self.add_qgeometry('path', {'prime_cpw_sub': prime_cpw},
                           width=p.prime_width + 2 * p.prime_gap,
                           subtract=True)
        self.add_qgeometry('path', {'second_cpw': second_cpw},
                           width=p.second_width,
                           fillet=p.fillet)
        self.add_qgeometry('path', {'second_cpw_sub': second_cpw_etch},
                           width=p.second_width + 2 * p.second_gap,
                           subtract=True,
                           fillet=p.fillet)

        #Add pins
        prime_pin_list = prime_cpw.coords
        second_pin_list = second_cpw.coords

        self.add_pin('prime_start',
                     points=np.array(prime_pin_list[::-1]),
                     width=p.prime_width,
                     input_as_norm=True)
        self.add_pin('prime_end',
                     points=np.array(prime_pin_list),
                     width=p.prime_width,
                     input_as_norm=True)
        self.add_pin('second_end',
                     points=np.array(second_pin_list[1:]),
                     width=p.second_width,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import BaseQubit


class TunableCoupler01(BaseQubit):
    """One of the tunable couplers Based off the implementation in
    https://arxiv.org/pdf/2011.01261.pdf.

    WIP - initial test structure

    Inherits `BaseQubit` class

    Description:
        Creates a tunable coupler, interdigitated capacitor to ground, with a junction to ground and a coupler arm.
        The shapes origin is shown with 0. X the location of the SQUID.

    ::

                            connection claw
                              _____
                X             |   |
        |       |       |     | | |     |       |
        |       |       |     | | |     |       | charge island
        |       |       |       |       |       |
        --------------------0--------------------

    .. image::
        TunableCoupler01.png

    .. meta::
        :description: Tunable Coupler 01

    Options:
        Convention: Values (unless noted) are strings with units included,
        (e.g., '30um')

    BaseQubit Default Options:
        * connection_pads: empty Dict -- Currently not used, connection count is static. (WIP)
        * _default_connection_pads: empty Dict -- The default values for the (if any) connection lines of the qubit.

    Default Options:
        * c_width: '400um' -- The width (x-axis) of the interdigitated charge island
        * l_width: '20um' -- The width of lines forming the body and arms of the charge island
        * l_gap: '10um' -- The dielectric gap of the charge island to ground
        * a_height: '60um' -- The length of the arms forming the 'fingers' of the charge island
        * cp_height: '15um' -- The thickness (y-axis) of the connection claw
        * cp_arm_length: '30um' -- The length of the 'fingers' of the connection claw (Warning: can break
          the component if they are too long)
        * cp_arm_width: '6um' -- The width of the 'fingers' of the connection claw (Warning: can break
          the component if too wide)
        * cp_gap: '6um' -- The dielectric gap of the connection claw
        * cp_gspace: '3um' -- How much ground remains between the connection claw and the charge island
        * fl_width: '5um' -- Width of the flux line
        * fl_gap: '3um' -- Dielectric gap of the flux line
        * fl_length: '10um' -- Length of the flux line for mutual inductance to the SQUID
        * fl_ground: '2um' -- Amount of ground between the SQUID and the flux line
        * _default_connection_pads: Currently empty
    """

    default_options = Dict(c_width='400um',
                           l_width='20um',
                           l_gap='10um',
                           a_height='60um',
                           cp_height='15um',
                           cp_arm_length='30um',
                           cp_arm_width='6um',
                           cp_gap='6um',
                           cp_gspace='3um',
                           fl_width='5um',
                           fl_gap='3um',
                           fl_length='10um',
                           fl_ground='2um')

    component_metadata = Dict(short_name='Pocket',
                              _qgeometry_table_path='True',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_junction='True')

    TOOLTIP = """One of the tunable couplers"""

    def make(self):
        """Builds the component."""
        p = self.p

        #Draw the charge island
        btm = draw.shapely.geometry.box(-p.c_width / 2, -p.l_width / 2, 0,
                                        p.l_width / 2)

        x_spot = p.c_width / 2 - p.l_width / 2

        arm1 = draw.shapely.geometry.box(-(x_spot + p.l_width / 2),
                                         p.l_width / 2,
                                         -(x_spot - p.l_width / 2), p.a_height)
        arm2 = draw.shapely.geometry.box(-((x_spot) * 3 / 5 + p.l_width / 2),
                                         p.l_width / 2,
                                         -((x_spot) * 3 / 5 - p.l_width / 2),
                                         p.a_height)
        arm3 = draw.shapely.geometry.box(-((x_spot) * 1 / 5 + p.l_width / 2),
                                         p.l_width / 2,
                                         -((x_spot) * 1 / 5 - p.l_width / 2),
                                         p.a_height)

        left_side = draw.shapely.ops.unary_union([btm, arm1, arm2, arm3])
        cap_island = draw.shapely.ops.unary_union([
            left_side,
            draw.shapely.affinity.scale(left_side,
                                        xfact=-1,
                                        yfact=1,
                                        origin=(0, 0))
        ])

        cap_subtract = cap_island.buffer(p.l_gap, cap_style=3, join_style=2)

        #Reference coordinates
        cpl_x = 1 / 5 * x_spot
        cpl_y = p.a_height + p.l_gap + p.cp_gap + p.cp_gspace
        fl_y = p.a_height + p.l_gap + p.fl_ground + p.fl_gap + p.fl_width / 2

        #Draw the junction and flux line
        rect_jj = draw.LineString([(-cpl_x * 3, p.a_height),
                                   (-cpl_x * 3, p.a_height + p.l_gap)])

        flux_line = draw.LineString([[-cpl_x * 3 - p.fl_length, fl_y],
                                     [-cpl_x * 3, fl_y],
                                     [-cpl_x * 3, fl_y + 0.01]])

        #Draw the connector
        cpl_x = 1 / 5 * x_spot
        cpl_y = p.a_height + p.l_gap + p.cp_gap + p.cp_gspace

        con_pad = draw.shapely.geometry.box(
            cpl_x - 1 / 5 * x_spot - p.cp_arm_width / 2, cpl_y,
            cpl_x + 1 / 5 * x_spot + p.cp_arm_width / 2, cpl_y + p.cp_height)

        con_arm_l = draw.shapely.geometry.box(
            cpl_x - 1 / 5 * x_spot - p.cp_arm_width / 2,
            cpl_y - p.cp_arm_length,
            cpl_x - 1 / 5 * x_spot + p.cp_arm_width / 2, cpl_y)

        con_arm_r = draw.shapely.geometry.box(
            cpl_x + 1 / 5 * x_spot - p.cp_arm_width / 2,
            cpl_y - p.cp_arm_length,
            cpl_x + 1 / 5 * x_spot + p.cp_arm_width / 2, cpl_y)

        con_body = draw.shapely.ops.unary_union([con_pad, con_arm_l, con_arm_r])

        con_sub = con_body.buffer(p.cp_gap, cap_style=3, join_style=2)

        con_pin = draw.LineString([[cpl_x, cpl_y], [cpl_x,
                                                    cpl_y + p.cp_height]])

        #Rotate and translate.
        c_items = [
            cap_island, cap_subtract, rect_jj, con_body, con_sub, flux_line,
            con_pin
        ]
        c_items = draw.rotate(c_items, p.orientation, origin=(0, 0))
        c_items = draw.translate(c_items, p.pos_x, p.pos_y)
        [
            cap_island, cap_subtract, rect_jj, con_body, con_sub, flux_line,
            con_pin
        ] = c_items

        #Add to qgeometry
        self.add_qgeometry('poly', {
            'cap_island': cap_island,
            'connector_body': con_body
        },
                           layer=p.layer)
        self.add_qgeometry('poly', {
            'cap_subtract': cap_subtract,
            'connector_sub': con_sub
        },
                           layer=p.layer,
                           subtract=True)

        self.add_qgeometry('path', {'flux_line': flux_line},
                           width=p.fl_width,
                           layer=p.layer)
        self.add_qgeometry('path', {'flux_line_sub': flux_line},
                           width=p.fl_width + 2 * p.fl_gap,
                           subtract=True,
                           layer=p.layer)

        self.add_qgeometry('junction', dict(rect_jj=rect_jj), width=p.l_width)

        #Add pin
        self.add_pin('Control',
                     points=np.array(con_pin.coords),
                     width=p.l_width,
                     input_as_norm=True)
        self.add_pin('Flux',
                     points=np.array(flux_line.coords[-2:]),
                     width=p.l_width,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent
import numpy as np


class NGon(QComponent):
    """A n-gon polygon. Eg. n = 3 : triangle, n = infinity : circle

    Inherits `QComponent` class.

    .. image::
        NGon.png

    .. meta::
        :description: N-Gon Polygon

    Default Options:
        Convention: Values (unless noted) are strings with units included,
        (e.g., '30um')

        * n: '3' -- Number of sides of the polygon
        * radius: '30um' -- The radius of the circle given n=infinity
        * subtract: 'False'
        * helper: 'False'
    """

    default_options = Dict(n='3',
                           radius='30um',
                           subtract='False',
                           helper='False')
    """Default drawing options"""

    TOOLTIP = """A n-gon polygon"""

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc."""
        p = self.p  # p for parsed parameters. Access to the parsed options.
        n = int(p.n)
        # Create the geometry
        # Generates a list of points
        n_polygon = [(p.radius * np.cos(2 * np.pi * x / n),
                      p.radius * np.sin(2 * np.pi * x / n)) for x in range(n)]
        # Converts said list into a shapely polygon
        n_polygon = draw.Polygon(n_polygon)

        n_polygon = draw.rotate(n_polygon, p.orientation, origin=(0, 0))
        n_polygon = draw.translate(n_polygon, p.pos_x, p.pos_y)

        ##############################################
        # add qgeometry
        self.add_qgeometry('poly', {'n_polygon': n_polygon},
                           subtract=p.subtract,
                           helper=p.helper,
                           layer=p.layer,
                           chip=p.chip)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Handles editing a QComponent."""

from typing import TYPE_CHECKING

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt, QTimer
from PySide6.QtWidgets import QTableView, QAbstractItemView

from ..bases.QWidget_PlaceholderText import QWidget_PlaceholderText

if TYPE_CHECKING:
    from ...main_window import MetalGUI


class QTableView_Options(QTableView, QWidget_PlaceholderText):
    """The class for QTableView_Options.

    This class inhertis the `QTableView` and `QWidget_PlaceholderText`
    classes.
    """

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Args:
            parent (QtWidgets.QWidget): parent widget
        """
        QTableView.__init__(self, parent)
        QWidget_PlaceholderText.__init__(self, "Select a QComponent to edit"\
            "\n\nfrom the QComponents window")
        QTimer.singleShot(
            200,
            self.style_me)  # not sure whu the ui isnt unpdating these here.

    def style_me(self):
        """Style the widget."""
        # Also can do in the ui file, but doesn't always transalte for me for some reason
        self.horizontalHeader().show()
        self.verticalHeader().hide()
        self.setAutoScroll(False)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

    #TODO: Maybe move to base class of utilty, along with the show template message
    def autoresize_columns(self, max_width: int = 200):
        """Rezie columsn to contents with maximim.

        Args:
            max (int): automatically resize the columns to the given size.  Defaults to 200.
        """
        self.resizeColumnsToContents()
        columns = self.model().columnCount()
        for i in range(columns):
            width = self.columnWidth(i)
            if width > max_width:
                self.setColumnWidth(i, max_width)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Module containing Net information storage."""
#from typing import Tuple
import pandas as pd
from qiskit_metal import logger


class QNet():
    """Use DataFrame to hold Net Information about the connected pins of a
    design.

    There is one unique net_id for each connected pin.
    """

    def __init__(self):
        """Hold the net information of all the USED pins within a design."""
        self.column_names = ['net_id', 'component_id', 'pin_name']
        self._net_info = pd.DataFrame(columns=self.column_names)
        self._qnet_latest_assigned_id = 0
        self.logger = logger  # type: logging.Logger

    def _get_new_net_id(self) -> int:
        """Provide unique new qnet_id.

        Returns:
            int: ID to use for storing a new net within _net_info.
        """
        self._qnet_latest_assigned_id += 1
        return self._qnet_latest_assigned_id

    @property
    def qnet_latest_assigned_id(self) -> int:
        """Return unique number for each net in table.

        Returns:
            int: For user of the design class to know the lastest id added to _net_info.
        """
        return self._qnet_latest_assigned_id

    @property
    def net_info(self) -> pd.DataFrame:
        """Provide table of all nets within the design.

        Returns:
            pd.DataFrame: Table of the net of pins within design.
        """
        return self._net_info

    def _check_arguments(self, comp1_id: int, pin1_name: str, comp2_id: int,
                         pin2_name: str) -> int:
        """Error check the arguments before using them.

        Args:
            comp1_id (int): Name of component 1.
            pin1_name (str): Corresponding pin name for component1.
            comp2_id (int): Name of component 2.
            pin2_name (str): Corresponding pin name for component2.

        Returns:
            int: 0 means should not be added to _net_info, else 1.
        """
        if not isinstance(comp1_id, int):
            self.logger.warning(
                f'Expected an int, but have {comp1_id}. '
                'The pins are were not entered to the net_info table.')
            return 0
        if not isinstance(comp2_id, int):
            self.logger.warning(
                f'Expected an int, but have {comp2_id}. '
                'The pins are were not entered to the net_info table.')
            return 0
        if not isinstance(pin1_name, str):
            self.logger.warning(
                f'Expected a string, but have {pin1_name}. '
                'The pins are were not entered to the net_info table.')
            return 0
        if not isinstance(pin2_name, str):
            self.logger.warning(
                f'Expected a string, but have {pin2_name}. '
                'The pins are were not entered to the net_info table.')
            return 0

        return 1

    def add_pins_to_table(self, comp1_id: int, pin1_name: str, comp2_id: int,
                          pin2_name: str) -> int:
        """Add two entries into the _net_info table. If either component/pin is
        already in net_info, the connection will NOT be added to the net_info.

        Args:
            comp1_id (int): Name of component 1.
            pin1_name (str): Corresponding pin name for component1.
            comp2_id (int): Name of component 2.
            pint2 (str): Corresponding pin name for component2.

        Returns:
            int: 0 if not added to list, otherwise the net_id
        """

        if self._check_arguments(comp1_id, pin1_name, comp2_id, pin2_name) == 0:
            return 0

        # Confirm the component-pin combination is NOT in _net_info, before adding them.
        for (net_identity, component_id,
             pin_name) in self._net_info.itertuples(index=False):
            if ((component_id == comp1_id) and (pin_name == pin1_name)):
                self.logger.warning(
                    f'Component: {comp1_id} and pin: {pin1_name} are '
                    f'already in net_info with net_id {net_identity}')
                return 0
            if ((component_id == comp2_id) and (pin_name == pin2_name)):
                self.logger.warning(
                    f'Component: {comp2_id} and pin: {pin2_name} are '
                    f'already in net_info with net_id {net_identity}')
                return 0

        net_id = self._get_new_net_id()

        entry1 = [net_id, comp1_id, pin1_name]
        entry2 = [net_id, comp2_id, pin2_name]
        temp_df = pd.DataFrame([entry1, entry2], columns=self.column_names)

        self._net_info = pd.concat([self._net_info, temp_df],
                                   axis=0,
                                   join='outer',
                                   ignore_index=True,
                                   sort=False,
                                   verify_integrity=False,
                                   copy=False)

        return net_id

    def delete_net_id(self, net_id_to_remove: int):
        """Removes the two entries with net_id_to_remove. If id is in
        _net_info, the entry will be removed.

        Args:
            net_id_to_remove (int): The id to remove.
        """

        self._net_info.drop(
            self._net_info.index[self._net_info['net_id'] == net_id_to_remove],
            inplace=True)

    def delete_all_pins_for_component(self, component_id_to_remove: int) -> set:
        """Delete all the pins for a given component id.

        Args:
            component_id_to_remove (int): Component ID to remove

        Returns:
            set: All deleted ids
        """
        all_net_id_deleted = set()

        for (net_identity, component_id,
             dummy_pin_name) in self._net_info.itertuples(index=False):
            if component_id == component_id_to_remove:
                all_net_id_deleted.add(net_identity)
                self.delete_net_id(net_identity)

        return all_net_id_deleted

    def get_components_and_pins_for_netid(self,
                                          net_id_search: int) -> pd.DataFrame:
        """Search with a net_id to get component id and pin name.

        Args:
            net_id_search (int): Unique net id which connects two pins within a design.

        Returns:
            pandas.DataFrame: Two rows of the net_info which have the same net_id_search.
        """
        df_subset_based_on_net_id = self._net_info[(
            self._net_info['net_id'] == net_id_search)]
        return df_subset_based_on_net_id



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Qiskit Metal main unit test functionality."""
import os
import fnmatch
import sys
import subprocess

if __name__ == '__main__':
    LIST_OF_FILES = os.listdir('.')
    PATTERN = "test*.py"
    print("====> Running the entire test suite now...")
    ERRORS_EXIST = 0

    # Open log files
    with open("test_results.log",
              "w") as results_log, open("test_errors.log", "w") as errors_log:
        for entry in LIST_OF_FILES:
            if fnmatch.fnmatch(entry, PATTERN):
                if entry != sys.argv[0]:
                    print(f"Running {entry} tests...")
                    results_log.write(f"Running {entry} tests...\n")

                    # Run the test and capture output
                    result = subprocess.run(f"python {entry}",
                                            shell=True,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE,
                                            text=True)

                    # Write the stdout and stderr to test_results.log
                    results_log.write(result.stdout)
                    results_log.write(result.stderr)
                    results_log.write("\n")

                    # If there's an error, write it to test_errors.log
                    if result.returncode != 0:
                        ERRORS_EXIST += 1
                        errors_log.write(f"Error in {entry}:\n")
                        errors_log.write(result.stderr)
                        errors_log.write("\n")

                    print("")

        print("All tests complete")
        results_log.write("All tests complete\n")
        if ERRORS_EXIST == 0:
            print("Congratulations, all the tests PASSED!")
            print("Have a nice day...")
            results_log.write("Congratulations, all the tests PASSED!\n")
            results_log.write("Have a nice day...\n")
        else:
            print("One or more tests have FAILED!")
            print("Scroll up to review the results")
            results_log.write("One or more tests have FAILED!\n")
            results_log.write("Scroll up to review the results\n")



#
from ast import parse
from qiskit_metal.renderers.renderer_ansys_pyaedt.hfss_renderer_aedt import QHFSSPyaedt
from qiskit_metal import Dict
from typing import List, Tuple, Union
import pandas as pd
import pyEPR as epr


class QHFSSEigenmodePyaedt(QHFSSPyaedt):
    """Subclass of pyaedt HFSS renderer for methods unique to driven-modal solutions within HFSS.
     QPyaedt Default Options:

    """
    name = 'aedt_hfss_eigenmode'

    default_setup = Dict(
        name="QHFSSEigenmodePyaedt_setup",
        MinimumFrequency="5.0",  # GHz
        NumModes="1",
        MaxDeltaFreq="0.5",
        MaximumPasses="10",
        MinimumPasses="1",
        MinimumConvergedPasses="1",
        PercentRefinement="30",
        BasisOrder="1")
    """aedt HFSS Options"""

    default_pyepr_options = Dict(
        ansys=Dict(dielectric_layers=[3],),
        hamiltonian=Dict(cos_trunc=7, fock_trunc=8, numeric=True),
        print_result=True,
    )
    """pyEPR options"""

    def __init__(self,
                 multilayer_design: 'MultiPlanar',
                 project_name: Union[str, None] = None,
                 design_name: Union[str, None] = None,
                 initiate=False,
                 options: Dict = None):
        """Create a QRenderer for HFSS simulations using pyaedt and multiplanar design.
        QHFSSPyaedt is subclassed from QPyaedt, subclassed from QRendererAnalysis and
        subclassed from QRenderer.  The default_setup options are expected to be defined by
        child class of QHFSSPyaedt for driven-modal and eigenmode solution types.


        Args:

            multilayer_design (MultiPlanar): Use QGeometry within MultiPlanar to obtain elements for Ansys.
            project_name (Union[str, None], optional): Give a name, or one will be made based on class name of renderer.
                                            Defaults to None.
            design_name (Union[str, None], optional): Give a name, or one will be made based on class name of renderer.
                                            Defaults to None.
            initiate (bool, optional): True to initiate the renderer. Defaults to False.
            options (Dict, optional):  Used to override all options. Defaults to None.
        """

        super().__init__(multilayer_design,
                         renderer_type='HFSS_EM',
                         project_name=project_name,
                         design_name=design_name,
                         initiate=initiate,
                         options=options)

        #make a class to read in pandas table.
        self.tables = None

    def add_hfss_em_setup(
            self,
            name: str = None,
            MinimumFrequency: float = None,  # GHz
            NumModes: int = None,
            MaxDeltaFreq: float = None,
            MaximumPasses: int = None,
            MinimumPasses: int = None,
            MinimumConvergedPasses: int = None,
            PercentRefinement: int = None,
            BasisOrder: int = None):
        """Create a solution setup in Ansys HFSS Driven-Modal solution type. If user does not provide
        arguments, they will be obtained from QHFSSDrivenmodalPyaedt.default_setup dict.

        Args:
            name (str, optional): _description_. Defaults to None.
            MinimumFrequency (float, optional):  Minimum frequency in GHz. Defaults to self.default_setup.
            NumModes (int, optional): Number of modes.  Defaults to self.default_setup.
            MaxDeltaFreq (float, optional):   Maximum difference in freq between consecutive passes.
                                        Defaults to self.default_setup.
            MaximumPasses (int, optional):  Maximum number of passes. Defaults to self.default_setup.
            MinimumPasses (int, optional): Minimum number of passes.Defaults to self.default_setup.
            MinimumConvergedPasses (int, optional): Minimum number of converged passes.
                                        Defaults to self.default_setup.
            PercentRefinement (int, optional): Percent refinement. Defaults to self.default_setup.
            BasisOrder (int, optional): Basis order. Defaults to self.default_setup.

        Returns:
            new_setup (pyaedt.modules.SolveSetup.SetupHFSS): pyAEDT simulation setup object.

        """

        self.activate_user_project_design()

        esu = self.default_setup

        if not name:
            name = self.parse_value(esu['name'])

        if name in self.current_app.setup_names:
            self.logger.warning(
                f'The setup name already exists within '
                f'project:{self.project_name} design: {self.design_name}. '
                f'So a new setup with name={name} was NOT added to design.')
            return

        if not MinimumFrequency:
            MinimumFrequency = float(self.parse_value(esu['MinimumFrequency']))
        if not NumModes:
            NumModes = int(self.parse_value(esu['NumModes']))
        if not MaxDeltaFreq:
            MaxDeltaFreq = float(self.parse_value(esu['MaxDeltaFreq']))
        if not MaximumPasses:
            MaximumPasses = int(self.parse_value(esu['MaximumPasses']))
        if not MinimumPasses:
            MinimumPasses = int(self.parse_value(esu['MinimumPasses']))
        if not MinimumConvergedPasses:
            MinimumConvergedPasses = int(
                self.parse_value(esu['MinimumConvergedPasses']))
        if not PercentRefinement:
            PercentRefinement = int(self.parse_value(esu['PercentRefinement']))
        if not BasisOrder:
            BasisOrder = int(self.parse_value(esu['BasisOrder']))

        new_setup = self.current_app.create_setup(name)

        new_setup.props['MinimumFrequency'] = f'{MinimumFrequency}GHz'
        new_setup.props['NumModes'] = NumModes
        new_setup.props['MaxDeltaFreq'] = MaxDeltaFreq
        new_setup.props['MaximumPasses'] = MaximumPasses
        new_setup.props['MinimumPasses'] = MinimumPasses
        new_setup.props['MinimumConvergedPasses'] = MinimumConvergedPasses
        new_setup.props['PercentRefinement'] = PercentRefinement
        new_setup.props['BasisOrder'] = BasisOrder

        new_setup.update()

        return new_setup

    def analyze_setup(self, setup_name: str) -> bool:
        """Run a specific solution setup in Ansys HFSS DrivenModal.

        Args:
            setup_name (str): Name of setup.

        Returns:
            bool: Value returned from pyaedt.analyze_setup().

        """
        # Activate project_name and design_name before anything else
        self.activate_user_project_design()

        if setup_name not in self.current_app.setup_names:
            self.logger.warning(
                f'Since the setup_name is not in the project/design which was used to start HFSS DrivenModal, '
                f'a new setup will be added to design with default settings for HFSS DrivenModal.'
            )
            self.add_hfss_dm_setup(setup_name)

        return self.current_app.analyze_setup(setup_name)

    def render_design(self,
                      selection: Union[list, None] = None,
                      open_pins: Union[list, None] = None,
                      port_list: Union[list, None] = None,
                      jj_to_port: Union[list, None] = None,
                      ignored_jjs: Union[list, None] = None,
                      box_plus_buffer: bool = True):
        """
        This render_design will add additional logic for just eigenmode design within project.

        Initiate rendering of components in design contained in selection,
        assuming they're valid. Components are rendered before the chips they
        reside on, and subtraction of negative shapes is performed at the very
        end.

        First obtain a list of IDs of components to render and a corresponding case, denoted by self.qcomp_ids
        and self.case, respectively. If self.case == 1, all components in QDesign are to be rendered.
        If self.case == 0, a strict subset of components in QDesign are to be rendered. Otherwise, if
        self.case == 2, one or more component names in selection cannot be found in QDesign.

        Chip_subtract_dict consists of component names (keys) and a set of all elements within each component that
        will eventually be subtracted from the ground plane. Add objects that are perfect conductors and/or have
        meshing to self.assign_perfE and self.assign_mesh, respectively; both are initialized as empty lists. Note
        that these objects are "refreshed" each time render_design is called (as opposed to in the init function)
        to clear QAnsysRenderer of any leftover items from the last call to render_design.

        Among the components selected for export, there may or may not be unused (unconnected) pins.
        The second parameter, open_pins, contains tuples of the form (component_name, pin_name) that
        specify exactly which pins should be open rather than shorted during the simulation. Both the
        component and pin name must be specified because the latter could be shared by multiple
        components. All pins in this list are rendered with an additional endcap in the form of a
        rectangular cutout, to be subtracted from its respective plane.

        The final parameter, box_plus_buffer, determines how the chip is drawn. When set to True, it takes the
        minimum rectangular bounding box of all rendered components and adds a buffer of x_buffer_width_mm and
        y_buffer_width_mm horizontally and vertically, respectively, to the chip size. The center of the chip
        lies at the midpoint x/y coordinates of the minimum rectangular bounding box and may change depending
        on which components are rendered and how they're positioned. If box generated with box_plus_buffer
        passes the chip size, the box will be cropped to chip size. If box_plus_buffer is False, however, the
        chip position and dimensions are taken from the chip info dictionary found in self.design, irrespective
        of what's being rendered. While this latter option is faster because it doesn't require calculating a
        bounding box, it runs the risk of rendered components being too close to the edge of the chip or even
        falling outside its boundaries.

        Args:
            selection (Union[list, None], optional): List of components to render. Defaults to None.
            open_pins (Union[list, None], optional): List of tuples of pins that are open. Defaults to None.
            box_plus_buffer (bool, optional): Either calculate a bounding box based on the location of rendered geometries
                                     or use chip size from design class.
        """
        # self.fill_info will hold the name of newly generated box,
        # along with information from layer stack
        self.fill_info = self.design.ls.get_layer_datatype_when_fill_is_true()

        # Draw in fill = True pieces. based on either full chip or box_plus

        # They are reset for each time render_design happens.

        # jj_to_port SHOULD not bew in eigenmode solution type. So will give error
        if jj_to_port or port_list:
            self.design.logger.error(
                f'In eigenmode solution, there should NOT be any values for jj_to_port or port_list.'
            )
            return

        super().render_design(selection, open_pins, port_list, jj_to_port,
                              ignored_jjs, box_plus_buffer)

        if self.case == 2:
            self.logger.warning(
                'Unable to proceed with rendering. Please check selection.')
            return

        self.activate_user_project_design()

        # Ansys default units is 'mm'?????, but metal using 'meter'.
        # pyaedt.generic.constants.SI_UNITS.Length is meter.

        # self.add_mesh()
        # self.assign_thin_conductor()
        # self.assign_nets()

        self.add_mesh()

        return

    def add_mesh(self):
        a = 5

    def should_render_junction(self, qgeom: pd.Series, port_list: Union[list,
                                                                        None],
                               jj_to_port: Union[list, None],
                               ignored_jjs: Union[list, None]) -> bool:
        """Logic Just for eigenmode

        Args:
            qgeom (pd.Series): One row of the junction table.
            port_list (Union[list, None]): Use definition from render_design.
            jj_to_port (Union[list, None]): Use definition from render_design.
            ignored_jjs (Union[list, None]): Use definition from render_design.

        Returns:
            bool: If the junction should be rendered based on row from junction
                table and ignored_jjs.
        """

        if ignored_jjs:
            search_junction = (self.design._components[qgeom["component"]].name,
                               qgeom['name'])
            if search_junction in ignored_jjs:
                return False

        return True

    def run_epr(self,
                dielectric_layers=None,
                cos_trunc: int = None,
                fock_trunc: int = None,
                numeric: bool = None,
                print_result: bool = None):
        """Run EPR analysis and return the results dict.

        Interpreting results:
        - ``f_0`` [MHz]: Eigenmode frequencies from HFSS (linear).
        - ``f_1`` [MHz]: Dressed mode frequencies (includes Lamb shift, etc.).
        - ``f_ND`` [MHz]: Numerically diagonalized frequencies.
        - ``chi_O1`` [MHz]: Analytic chis (cosine truncated to 4th order, first-order perturbation).
        - ``chi_ND`` [MHz]: Numerically diagonalized chi matrix (diag = anharmonicity, off-diag = cross-Kerr).

        Args:
            dielectric_layers: Layers treated as dielectrics (defaults to ``default_pyepr_options``).
            cos_trunc: Cosine truncation (defaults to ``default_pyepr_options``).
            fock_trunc: Fock truncation (defaults to ``default_pyepr_options``).
            numeric: Use numerical diagonalization (defaults to ``default_pyepr_options``).
            print_result: Print results (defaults to ``default_pyepr_options``).

        Returns:
            dict: ``self.epr_quantum_analysis.data`` with all EPR results.
        """
        if (print_result == None):
            print_result = self.default_pyepr_options.print_result

        # Sets ANSYS to project associated with self.design
        self.activate_user_project_design()

        # Connect EPR to ANSYS
        self.pinfo = epr.ProjectInfo()

        # Tells pyEPR where Johsephson Junctions are located in ANSYS
        self.setup_jjs_for_epr()

        # Tells pyEPR which components have dissipative elements
        self.setup_dielectric_for_epr(dielectric_layers)

        # Executes the EPR analysis
        self.epr_distributed_analysis = epr.DistributedAnalysis(self.pinfo)
        self.epr_distributed_analysis.do_EPR_analysis()

        # Find observable quantities from energy-partipation ratios
        self.epr_spectrum_analysis(cos_trunc=cos_trunc,
                                   fock_trunc=fock_trunc,
                                   print_result=print_result)

        # Print results?
        if print_result:
            self.epr_quantum_analysis.report_results(numeric=numeric)

        # Release ANSYS from pyEPR script
        self.pinfo.disconnect()

        return self.epr_quantum_analysis.data

    def setup_jjs_for_epr(self):
        """
        Finds all names, inductances, and capacitances of Josephson Junctions rendered into ANSYS.
        """
        # Get all josephson junctions from rendered components table
        geom_table = self.path_poly_and_junction_with_valid_comps
        all_jjs = geom_table.loc[geom_table['name'].str.contains('rect_jj')]
        all_jjs = all_jjs.reset_index(drop=True)

        for i, row in all_jjs.iterrows():
            ### Parsing Data ###
            component = str(row['component'])
            name = str(row['name'])
            inductance = row['aedt_hfss_inductance']  # Lj in Henries
            capacitance = row['aedt_hfss_capacitance']  # Cj in Farads

            # Get ANSYS > Model > Sheet corresponding to JJs
            rect_name = 'JJ_rect_Lj_' + component + '_' + name

            # Get ANSYS > Model > Lines corresponding to JJs
            line_name = 'JJ_Lj_' + component + '_' + name + '_'

            ### Appending data ###
            # Add global Lj and Cj variables to ANSYS (for EPR analysis)
            ansys_Lj_name = f'Lj_{i}'
            ansys_Cj_name = f'Cj_{i}'

            self.set_variable(ansys_Lj_name, str(inductance * 1E9) + 'nH')
            self.set_variable(ansys_Cj_name, str(capacitance * 1E15) + 'fF')

            # Append data in pyEPR.ProjectInfo.junctions data format
            junction_dict = {
                'Lj_variable': ansys_Lj_name,
                'rect': rect_name,
                'line': line_name,
                'length': epr.parse_units('100um'),
                'Cj_variable': ansys_Cj_name
            }

            self.pinfo.junctions[f'j{i}'] = junction_dict

        self.pinfo.validate_junction_info()

    def setup_dielectric_for_epr(self, dielectric_layers=None):
        """
        Find name of dielectric layer rendered in ANSYS, then 
        define it as a dissipative dielectric surface for pyEPR.

        Args:
            dielectric_layers (list, optional): Specify which layers are dielectrics.
                Layer specified in `LayerStackHandler.ls_df['layer']`.
                Defaults to self.default_pyepr_options, which is the default silicon layer.

        """
        # Check for default values
        if (dielectric_layers == None):
            dielectric_layers = self.default_pyepr_options.ansys.dielectric_layers

        # Check if layerstack (self.design.ls) is uniquely specified
        ls_unique = self.design.ls.is_layer_data_unique()
        if (ls_unique != True):
            raise ValueError('Layer data in `MultiPlanar` design is not unique')

        dielectric_names = []
        ls_df = self.design.ls.ls_df
        for layer in dielectric_layers:
            # Find layer names
            selected_ls_df = ls_df[ls_df['layer'] == layer]
            datatype = selected_ls_df['datatype'].values[0]
            dielectric_name = f'layer_{layer}_datatype_{datatype}_plane'

            dielectric_names.append(dielectric_name)

        # Define them as dielectrics in pyEPR.ProjectInfo object
        self.pinfo.dissipative['dielectric_surfaces'] = dielectric_names

    def epr_spectrum_analysis(self,
                              cos_trunc: int = None,
                              fock_trunc: int = None,
                              print_result: bool = None):
        """Core EPR analysis method.

        Args:
            cos_trunc (int, optional): Truncation of the cosine. Defaults to self.default_pyepr_options.
            fock_trunc (int, optional): Truncation of the fock. Defaults to self.default_pyepr_options.
            print (boo, optional): Print results of EPR analysis. Defaults to self.default_pyepr_options.
        """
        if (cos_trunc == None):
            cos_trunc = self.default_pyepr_options.hamiltonian.cos_trunc
        if (fock_trunc == None):
            fock_trunc = self.default_pyepr_options.hamiltonian.fock_trunc
        if (print_result == None):
            print_result = self.default_pyepr_options.print_result

        self.epr_quantum_analysis = epr.QuantumAnalysis(
            self.epr_distributed_analysis.data_filename)
        self.epr_quantum_analysis.analyze_all_variations(
            cos_trunc=cos_trunc,
            fock_trunc=fock_trunc,
            print_result=print_result)

    def epr_report_hamiltonian(self, numeric=None):
        """Reports in a markdown friendly table the hamiltonian results.

        Args:
            numeric (bool, optional): Use numerical diagonalization. Defaults to self.default_pyepr_options.
        """
        if (numeric == None):
            numeric = self.default_pyepr_options.hamiltonian.numeric

        self.epr_quantum_analysis.plot_hamiltonian_results()
        self.epr_quantum_analysis.report_results(numeric=numeric)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6 import QtWidgets, QtCore, QtGui
from PySide6.QtGui import QContextMenuEvent
from PySide6.QtCore import QPoint, QModelIndex, QTimer
from PySide6.QtWidgets import QInputDialog, QLineEdit, QTableView, QMenu, QMessageBox
from PySide6.QtWidgets import QAbstractItemView


class RightClickView(QTableView):
    """Standard QTableView with drop-down context menu upon right-clicking.
    Menu allows for row deletion and renaming a cell.

    This class extends the `QTableView` class.

    Access:
        gui.variables_window.ui.tableView
    """

    def __init__(self, parent):
        """Provide access to GUI QMainWindow via parent."""
        super().__init__(parent)
        self.gui = parent.parent()  # this is not the main gui

        QTimer.singleShot(
            200,
            self.style_me)  # not sure whu the ui isnt unpdating these here.

    def style_me(self):
        """Style this widget."""
        # Also can do in the ui file, but doesn't always transalte for me for some reason
        self.horizontalHeader().show()
        self.verticalHeader().hide()
        self.setAutoScroll(False)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

    def contextMenuEvent(self, event: QContextMenuEvent):
        """Create options for drop-down context menu.

        Args:
            event (QContextMenuEvent): The event
        """
        self.right_click_menu = QMenu(self)
        self.right_click_menu._d = self.right_click_menu.addAction('Delete')
        self.right_click_menu._r = self.right_click_menu.addAction('Rename')
        row_name, index = self.getPosition(event.pos())
        self.right_click_menu._d.triggered.connect(
            lambda: self.deleteRow(row_name, index.row()))
        self.right_click_menu._r.triggered.connect(
            lambda: self.renameRow(row_name, index))
        self.right_click_menu.action = self.right_click_menu.exec_(
            self.mapToGlobal(event.pos()))

    def getPosition(self, clickedIndex: QPoint):
        """Obtain location of clicked cell in form of row name and number.

        Args:
            clickedIndex (QPoint): The QPoint of the click

        Returns:
            tuple: name, index
        """
        index = self.indexAt(clickedIndex)
        name = list(self.gui.model._data.keys())[index.row()]
        if index.isValid():
            return name, index
        return None, None

    def deleteRow(self, row_name: str, row_number: int):
        """Create message box to confirm row deletion.

        Args:
            row_name (str): Name of the row to delete
            row_number (int): Number of the row being deleted
        """
        if row_number > -1:
            choice = QMessageBox.question(
                self, '', f'Are you sure you want to delete {row_name}?',
                QMessageBox.Yes | QMessageBox.No)
            if choice == QMessageBox.Yes:
                model = self.model()
                model.removeRows(row_number)

    def renameRow(self, row_name: str, index: QModelIndex):
        """Create message box to confirm row renaming and new name.

        Args:
            row_name (str): Name of the row to rename
            index (QModelIndex): Index of the row to rename
        """
        if index.row() > -1:
            text, okPressed = QInputDialog.getText(self, 'Rename',
                                                   f'Rename {row_name} to:',
                                                   QLineEdit.Normal, '')
            if okPressed and (text != ''):
                model = self.model()
                model.setData(index, text)
